import cv2
import numpy as np
from numpy.typing import NDArray

from autowsgr.constants.image_templates import MyTemplate


def relative_to_absolute(record_pos, resolution=(960, 540)):
    """将相对坐标转换为绝对坐标"""
    rel_x, rel_y = record_pos
    _w, _h = resolution
    assert 0 <= rel_x <= 1 and 0 <= rel_y <= 1, 'rel_x and rel_y should be in [0, 1]'
    abs_x = _w * rel_x
    abs_y = _h * rel_y
    return abs_x, abs_y


def cv_show_image(img):
    """调试用, 展示图像, 按任意键关闭窗口

    img: numpy.ndarray 格式的图片
    """
    cv2.imshow('Image', img)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


def absolute_to_relative(absolute_pos, resolution=(960, 540)) -> tuple[float, float]:
    """将绝对坐标转换为相对坐标"""
    abs_x, abs_y = absolute_pos
    _w, _h = resolution
    assert 0 <= abs_x <= _w and 0 <= abs_y <= _h, f'{abs_x, abs_y} should be in {resolution}'
    rel_x = abs_x / _w
    rel_y = abs_y / _h
    return rel_x, rel_y


# https://github.com/leimao/Rotated-Rectangle-Crop-OpenCV/blob/master/rotated_rect_crop.py
def inside_rect(rect, num_cols, num_rows):
    # Determine if the four corners of the rectangle are inside the rectangle with width and height
    # rect tuple
    # center (x,y), (width, height), angle of rotation (to the row)
    # center  The rectangle mass center.
    # center tuple (x, y): x is regarding to the width (number of columns) of the image, y is regarding to the height (number of rows) of the image.
    # size    Width and height of the rectangle.
    # angle   The rotation angle in a clockwise direction. When the angle is 0, 90, 180, 270 etc., the rectangle becomes an up-right rectangle.
    # Return:
    # True: if the rotated sub rectangle is side the up-right rectangle
    # False: else

    rect_center = rect[0]
    rect_center_x = rect_center[0]
    rect_center_y = rect_center[1]

    if (rect_center_x < 0) or (rect_center_x > num_cols):
        return False
    if (rect_center_y < 0) or (rect_center_y > num_rows):
        return False

    # https://docs.opencv.org/3.0-beta/modules/imgproc/doc/structural_analysis_and_shape_descriptors.html
    box = cv2.boxPoints(rect)

    x_max = int(np.max(box[:, 0]))
    x_min = int(np.min(box[:, 0]))
    y_max = int(np.max(box[:, 1]))
    y_min = int(np.min(box[:, 1]))

    return (x_max <= num_cols) and (x_min >= 0) and (y_max <= num_rows) and (y_min >= 0)


def rect_bbx(rect):
    # Rectangle bounding box for rotated rectangle
    # Example:
    # rotated rectangle: height 4, width 4, center (10, 10), angle 45 degree
    # bounding box for this rotated rectangle, height 4*sqrt(2), width 4*sqrt(2), center (10, 10), angle 0 degree

    box = cv2.boxPoints(rect)

    x_max = int(np.max(box[:, 0]))
    x_min = int(np.min(box[:, 0]))
    y_max = int(np.max(box[:, 1]))
    y_min = int(np.min(box[:, 1]))

    # Top-left
    # (x_min, y_min)
    # Top-right
    # (x_min, y_max)
    # Bottom-left
    #  (x_max, y_min)
    # Bottom-right
    # (x_max, y_max)
    # Width
    # y_max - y_min
    # Height
    # x_max - x_min
    # Center
    # (x_min + x_max) // 2, (y_min + y_max) // 2

    center = (int((x_min + x_max) // 2), int((y_min + y_max) // 2))
    width = int(x_max - x_min)
    height = int(y_max - y_min)
    angle = 0

    return (center, (width, height), angle)


def image_rotate_without_crop(mat, angle):
    # https://stackoverflow.com/questions/22041699/rotate-an-image-without-cropping-in-opencv-in-c
    # angle in degrees

    height, width = mat.shape[:2]
    image_center = (width / 2, height / 2)

    rotation_mat = cv2.getRotationMatrix2D(image_center, angle, 1)

    abs_cos = abs(rotation_mat[0, 0])
    abs_sin = abs(rotation_mat[0, 1])

    bound_w = int(height * abs_sin + width * abs_cos)
    bound_h = int(height * abs_cos + width * abs_sin)

    rotation_mat[0, 2] += bound_w / 2 - image_center[0]
    rotation_mat[1, 2] += bound_h / 2 - image_center[1]

    return cv2.warpAffine(mat, rotation_mat, (bound_w, bound_h))


def crop_rectangle_relative(
    image,
    x_ratio: float,
    y_ratio: float,
    width_ratio: float,
    height_ratio: float,
):
    """
    根据相对坐标和尺寸裁剪图像。

    参数:
    image -- 输入的 numpy 数组格式的图像。image[row][col]
    x_ratio -- 左上角 x 坐标的相对位置
    y_ratio -- 左上角 y 坐标的相对位置
    width_ratio -- 裁剪区域的宽度相对图像宽度的比例
    height_ratio -- 裁剪区域的高度相对图像高度的比例

    返回:
    cropped_image -- 裁剪后的图像。
    """
    # 获取图像的尺寸
    height, width = image.shape[:2]

    # 计算裁剪区域的左上角坐标和尺寸
    start_x = int(width * x_ratio)
    start_y = int(height * y_ratio)
    end_x = start_x + int(width * width_ratio)
    end_y = start_y + int(height * height_ratio)

    # 裁剪图像
    return image[start_y:end_y, start_x:end_x]


def crop_rectangle(image, rect):
    # rect has to be upright

    num_rows = image.shape[0]
    num_cols = image.shape[1]

    if not inside_rect(rect=rect, num_cols=num_cols, num_rows=num_rows):
        print('Proposed rectangle is not fully in the image.')
        return None

    rect_center = rect[0]
    rect_center_x = rect_center[0]
    rect_center_y = rect_center[1]
    rect_width = rect[1][0]
    rect_height = rect[1][1]

    return image[
        rect_center_y - rect_height // 2 : rect_center_y + rect_height - rect_height // 2,
        rect_center_x - rect_width // 2 : rect_center_x + rect_width - rect_width // 2,
    ]


def crop_rotated_rectangle(image, rect):
    # Crop a rotated rectangle from a image

    num_rows = image.shape[0]
    num_cols = image.shape[1]

    if not inside_rect(rect=rect, num_cols=num_cols, num_rows=num_rows):
        print('Proposed rectangle is not fully in the image.')
        return None

    rotated_angle = rect[2]

    rect_bbx_upright = rect_bbx(rect=rect)
    rect_bbx_upright_image = crop_rectangle(image=image, rect=rect_bbx_upright)

    rotated_rect_bbx_upright_image = image_rotate_without_crop(
        mat=rect_bbx_upright_image,
        angle=rotated_angle,
    )

    rect_width = rect[1][0]
    rect_height = rect[1][1]

    crop_center = (
        rotated_rect_bbx_upright_image.shape[1] // 2,
        rotated_rect_bbx_upright_image.shape[0] // 2,
    )

    return rotated_rect_bbx_upright_image[
        crop_center[1] - rect_height // 2 : crop_center[1] + (rect_height - rect_height // 2),
        crop_center[0] - rect_width // 2 : crop_center[0] + (rect_width - rect_width // 2),
    ]


def crop_image(
    image: NDArray,
    pos1: tuple[float, float],
    pos2: tuple[float, float],
    rotation: int = 0,
    debug: bool = False,
):
    """裁剪出矩形, pos1 左下角相对位置, pos2 右上角相对位置。如果指定旋转角度，则返回剪裁+旋转后的图片

    Args:
        image (NDArray): 图片
        pos1 (Tuple[float, float]): 左下角相对位置
        pos2 (Tuple[float, float]): 右上角相对位置
        rotation (int, optional): 旋转角度[-180, 180]. 正值顺时针旋转. Defaults to 0.
        debug (bool, optional): 是否保存调试图片. Defaults to False.
    """
    resolution = (image.shape[1], image.shape[0])
    x1, y2 = map(int, relative_to_absolute(pos1, resolution))
    x2, y1 = map(int, relative_to_absolute(pos2, resolution))

    if rotation == 0:
        ret = image[y1:y2, x1:x2]
    else:
        # 矩形的对角线长度，以及对角线角度
        diag = ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5
        diag_orientation = np.arctan2(y2 - y1, x2 - x1)
        diag_angle = diag_orientation - np.radians(rotation)
        # 计算矩形长、宽
        width = int(diag * np.cos(diag_angle))
        height = int(diag * np.sin(diag_angle))
        # 结构化表示矩形
        rect = ((x1 + x2) // 2, (y1 + y2) // 2), (width, height), 360 - rotation
        ret = crop_rotated_rectangle(image, rect)

    if debug:
        if ret is None:
            return None
        cv2.imwrite('crop_image.png', ret)

    return ret


def locate_image_center(
    image: NDArray,
    query: MyTemplate,
    confidence=0.85,
):
    """从原图像中尝试找出一个置信度相对于模板图像最高的矩阵区域的中心坐标

    Args:
        image (np.ndarray): 原图像
        query (MyTemplate): 模板图像
        confidence (float, optional): 置信度阈值. Defaults to 0.85.

    Returns:
        如果匹配结果中有超过阈值的,返回置信度最高的结果的中心绝对坐标:Tuple(int,int)

        否则返回 None
    """
    query.threshold = confidence
    return query.match_in(image)


def match_nearest_index(
    pos: tuple[int, int],
    positions: list[tuple[int, int]],
    metric: str = 'l2',
):
    """找出离目标点最近的点的索引

    Args:
        pos (Tuple[int, int]): 目标点
        positions (List[Tuple[int, int]]): 点的列表
        metric (str, optional): 距离度量方式. Defaults to "l2".
    Returns:
        int: 最近点的索引
    """
    min_distance = float('inf')
    min_index = -1
    for i, p in enumerate(positions):
        if metric == 'l2':
            distance = (p[0] - pos[0]) ** 2 + (p[1] - pos[1]) ** 2
        elif metric == 'l1':
            distance = abs(p[0] - pos[0]) + abs(p[1] - pos[1])
        else:
            raise ValueError('unsupported metric ' + str(metric))
        if distance < min_distance:
            min_distance = distance
            min_index = i
    return min_index



import copy
import time
from typing_extensions import Protocol

from autowsgr.configs import NodeConfig
from autowsgr.constants.custom_exceptions import ImageNotFoundErr, NetworkErr
from autowsgr.constants.image_templates import IMG, MyTemplate
from autowsgr.constants.other_constants import ALL_SHIP_TYPES
from autowsgr.constants.positions import BLOOD_BAR_POSITION
from autowsgr.constants.ui import Node
from autowsgr.game.expedition import Expedition
from autowsgr.game.game_operation import (
    click_result,
    destroy_ship,
    detect_ship_stats,
    get_ship,
    match_night,
)
from autowsgr.game.get_game_info import get_enemy_condition, get_enemy_formation
from autowsgr.timer import Timer
from autowsgr.types import ConditionFlag, Formation, SearchEnemyAction
from autowsgr.utils.logger import Logger
from autowsgr.utils.math_functions import get_nearest


def start_march(timer: Timer, position: tuple[int, int] = (900, 500)) -> ConditionFlag:
    timer.click(*position, 1, delay=0)
    start_time = time.time()
    while timer.identify_page('fight_prepare_page'):
        if time.time() - start_time > 3:
            timer.click(*position, 1, delay=0)
            time.sleep(1)
        if timer.image_exist(IMG.symbol_image[3], need_screen_shot=False):
            return ConditionFlag.DOCK_FULL
        if timer.image_exist(IMG.symbol_image[9], need_screen_shot=False, confidence=0.8):
            time.sleep(1)
            return ConditionFlag.BATTLE_TIMES_EXCEED
        if time.time() - start_time > 15:
            if timer.process_bad_network():
                if timer.identify_page('fight_prepare_page'):
                    return start_march(timer, position)
                NetworkErr('stats unknown')
            else:
                raise TimeoutError('map_fight prepare timeout')
    return ConditionFlag.OPERATION_SUCCESS


class FightResultInfo:
    def __init__(self, timer: Timer, ship_stats) -> None:
        try:
            mvp_pos = timer.wait_image(IMG.fight_image[14])
            self.mvp = get_nearest((mvp_pos[0], mvp_pos[1] + 20), BLOOD_BAR_POSITION[1])
        except Exception as e:
            timer.log_screen(name='mvp_image')
            timer.logger.warning(f"can't identify mvp, error: {e}")
        self.ship_stats = detect_ship_stats(timer, 'sumup', ship_stats)

        self.result = timer.wait_images(IMG.fight_result, timeout=5)
        if timer.image_exist(IMG.fight_result['SS'], need_screen_shot=False):
            self.result = 'SS'
        if self.result is None:
            timer.log_screen()
            timer.logger.warning("can't identify fight result, screen logged")

    def __str__(self) -> str:
        return f'MVP 为 {self.mvp} 号位, 战果为 {self.result}'

    def __lt__(self, other) -> bool:  # <
        order = ['D', 'C', 'B', 'A', 'S', 'SS']
        if isinstance(other, FightResultInfo):
            other = other.result

        return order.index(self.result) < order.index(other)

    def __le__(self, other) -> bool:  # <=
        order = ['D', 'C', 'B', 'A', 'S', 'SS']
        if isinstance(other, FightResultInfo):
            other = other.result

        return order.index(self.result) <= order.index(other)

    def __gt__(self, other) -> bool:  # >
        return not (self <= other)

    def __ge__(self, other) -> bool:  # >=
        return not (self < other)


class FightEvent:
    """战斗事件类
    事件列表: 战况选择, 获取资源, 索敌成功, 迂回, 阵型选择, 进入战斗, 是否夜战, 战斗结算, 获取舰船, 继续前进, 自动回港

    状态: 一个字典, 一共三种键值
        position: 位置, 所有事件均存在

        ship_stats: 我方舰船状态(仅在 "继续前进" 事件存在)

        enemies: 敌方舰船(仅在 "索敌成功" 事件存在), 字典或 "索敌失败"

        info: 其它额外信息(仅在 "自动回港" 事件存在)

    动作: 一个字符串
        "继续": 获取资源, 迂回, 战斗结算, 获取舰船, 自动回港等不需要决策的操作

        数字字符串: 战况选择的决策

        "SL"/数字字符串: 阵型选择的决策

        "继续"/ "SL": 进入战斗后的决策

        "战斗"/"撤退"/"迂回": 索敌成功的决策

        "追击"/"撤退": 夜战的决策

        "回港/前进": 是否前进的选择(战斗结算完毕后)

    结果: 一个字符串
        "无": 战况选择, 获取资源, 索敌成功, 阵型选择, 进入战斗, 是否夜战, 继续前进, 自动回港,

        (FightResultInfo): 表示战果信息, 战果结算

        舰船名: 获取舰船
    """

    def __init__(self, event, stats, action='继续', result='无') -> None:
        self.event = event
        self.stats = stats
        self.action = action
        self.result = result

    def __str__(self) -> str:
        return f'事件:{self.event}, 状态:{self.stats}, 动作:{self.action}, 结果:{self.result!s}'

    def __repr__(self) -> str:
        return f'FightEvent({self.event}, {self.stats}, {self.action}, {self.result})'


class FightHistory:
    """记录并处理战斗历史信息"""

    def __init__(self) -> None:
        self.events: list[FightEvent] = []

    def add_event(self, event, point, action='继续', result='无'):
        self.events.append(FightEvent(event, point, action, result))

    def reset(self):
        self.events = []

    def get_fight_results(self):
        results_dict = {}
        results_list = []
        for event in self.events:
            if event.event == '战果结算':
                if event.stats['position'].isalpha():
                    results_dict[event.stats['position']] = event.result
                else:
                    results_list.append(event.result)
        return results_list if len(results_list) else results_dict

    def get_last_point(self):
        return self.events[-1].stats['position']

    def __str__(self) -> str:
        return ''.join(str(event) + '\n' for event in self.events)


class FightInfo(Protocol):
    """存储战斗中需要用到的所有状态信息, 以及更新逻辑"""

    timer: Timer
    logger: Logger
    oil: int
    ammo: int
    enemies: dict
    ship_stats: list
    fight_history: FightHistory
    enemy_formation: str
    node: str

    # =============== 静态属性 ===============
    end_page: str
    """结束战斗后的页面，用于判断是否已经返回战斗外界面"""
    successor_states: dict[str, list[str] | dict[str, list[str]]]
    """战斗流程的有向图建模。格式为:
        1. {状态名: [后继状态1, 后继状态2, ...]}
        2. {状态名: {动作1: [后继状态1, 后继状态2, ...], 动作2: [后继状态1, 后继状态2, ...]}}
    """
    state2image: dict[str, list[MyTemplate, float]]
    """所需用到的图片模板。格式为 {状态明： [模板，等待时间(秒)]}"""
    after_match_delay: dict[str, float]
    """匹配成功后的延时。格式为 {状态名: 延时时间(秒)}"""

    # =============== 运行时属性 ===============
    last_state: str
    """上一个状态 s_{t-1}"""
    last_action: str
    """上一个动作 a_{t-1}"""
    state: str
    """当前状态 s_t"""

    def __init__(self, timer: Timer) -> None:
        self.timer = timer
        self.logger = timer.logger

        self.last_state = ''
        self.last_action = ''
        self.state = ''

        self.enemies = {}  # 敌方舰船列表
        self.ship_stats = []  # 我方舰船血量列表
        self.oil = 10  # 我方剩余油量
        self.ammo = 10  # 我方剩余弹药量
        self.fight_history = FightHistory()  # 战斗结果记录

    def update_state(self):
        self.last_state = self.state

        # 计算当前可能的状态
        possible_states = copy.deepcopy(self.successor_states[self.last_state])
        if isinstance(possible_states, dict):
            possible_states = possible_states[self.last_action]
        modified_timeout = [-1 for _ in possible_states]  # 某些状态需要修改等待时间
        for i, state in enumerate(possible_states):
            if isinstance(state, list):
                state, timeout = state
                possible_states[i] = state
                modified_timeout[i] = timeout
        if self.timer.config.show_match_fight_stage:
            self.logger.debug('waiting:', possible_states, '  ')
        images = [self.state2image[state][0] for state in possible_states]
        timeout = [self.state2image[state][1] for state in possible_states]
        confidence = min(
            [0.8]
            + [
                self.state2image[state][2]
                for state in possible_states
                if len(self.state2image[state]) >= 3
            ],
        )
        timeout = [
            timeout[i] if modified_timeout[i] == -1 else modified_timeout[i]
            for i in range(len(timeout))
        ]
        timeout = max(timeout)
        # 等待其中一种出现
        fun_start_time = time.time()
        while time.time() - fun_start_time <= timeout:
            self._before_match()

            # 尝试匹配
            ret = [self.timer.image_exist(image, False, confidence=confidence) for image in images]
            if any(ret):
                self.state = possible_states[ret.index(True)]
                # 查询是否有匹配后延时
                if self.state in self.after_match_delay:
                    delay = self.after_match_delay[self.state]
                    time.sleep(delay)

                if self.timer.config.show_match_fight_stage:
                    self.logger.info(f'matched: {self.state}')
                self._after_match()

                return self.state

        # 匹配不到时报错
        self.logger.warning(
            f'匹配状态失败! state: {self.state}  last_action: {self.last_action}',
        )
        self.timer.log_screen(True)
        for image in images:
            self.logger.log_image(image, f'match_{time.time()!s}.PNG')
        raise ImageNotFoundErr

    def _before_match(self):
        """每一轮尝试匹配状态前执行的操作"""

    def _after_match(self):
        """匹配到状态后执行的操作"""
        if self.state == 'spot_enemy_success':
            self.enemies = get_enemy_condition(self.timer, 'fight')
            self.enemy_formation = get_enemy_formation(self.timer)
        if self.state == 'result':
            try:
                result = FightResultInfo(self.timer, self.ship_stats)
                self.ship_stats = result.ship_stats
                self.fight_history.add_event(
                    '战果结算',
                    {
                        'position': (
                            self.node
                            if 'node' in self.__dict__
                            else f'此类战斗({type(self)})不支持节点信息'
                        ),
                    },
                    result=result,
                )
            except Exception as e:
                self.logger.warning(f'战果结算记录失败：{e}')

    def reset(self):
        """需要记录与初始化的战斗信息"""


class FightPlan(Protocol):
    info: FightInfo
    timer: Timer
    logger: Logger
    fight_logs: list[FightHistory]

    def __init__(self, timer: Timer) -> None:
        # 把 timer 引用作为内置对象，减少函数调用的时候所需传入的参数
        self.timer = timer
        self.logger = timer.logger
        self.fight_logs = []

    def fight(self) -> ConditionFlag:
        self.info.reset()  # 初始化战斗信息
        while True:
            ret = self._make_decision()
            if ret == ConditionFlag.FIGHT_CONTINUE:
                continue
            if ret == ConditionFlag.SL:
                self._sl()
                return ConditionFlag.SL
            if ret == ConditionFlag.FIGHT_END:
                self.timer.set_page(self.info.end_page)
                self.fight_logs.append(self.info.fight_history)
                return ConditionFlag.OPERATION_SUCCESS

    def run_for_times(self, times, gap=1800) -> ConditionFlag:
        """多次执行同一任务, 自动进行远征操作
        Args:
            times (int): 任务执行总次数

            gap (int): 强制远征检查的间隔时间
        Raise:
            RuntimeError: 战斗进行时出现错误
        Returns:
            ConditionFlag
        """
        assert times >= 1
        expedition = Expedition(self.timer)
        self.timer.goto_game_page('map_page')
        for i in range(times):
            if time.time() - self.timer.last_expedition_check_time >= gap:
                expedition.run(True)
            elif isinstance(self.timer.now_page, Node) and self.timer.now_page.name == 'map_page':
                expedition.run(False)
                self.timer.goto_game_page('map_page')
            fight_flag = self.run()
            if fight_flag not in [ConditionFlag.SL, ConditionFlag.OPERATION_SUCCESS]:
                if fight_flag == ConditionFlag.DOCK_FULL:
                    return ConditionFlag.DOCK_FULL
                if fight_flag == ConditionFlag.SKIP_FIGHT:
                    return ConditionFlag.SKIP_FIGHT
                raise RuntimeError(f'战斗进行时出现异常, 信息为 {fight_flag}')
            self.timer.logger.info(f'已出击次数:{i+1}，目标次数{times}')
        return ConditionFlag.OPERATION_SUCCESS

    def run(self, retry_times=0, max_try_times=5) -> ConditionFlag:
        """主函数，负责一次完整的战斗.
        Args:
            retry_times (int): 重试次数
            max_try_times (int): 最大尝试次数
        Returns:
            ConditionFlag
        """
        # 战斗前逻辑
        ret = self._enter_fight()

        if ret == ConditionFlag.OPERATION_SUCCESS:
            pass
        elif ret == ConditionFlag.DOCK_FULL:
            # 自动解装功能
            if self.timer.config.dock_full_destroy and retry_times < max_try_times:
                self.logger.debug(f'船坞已满, 正在解装, 尝试次数:{retry_times+1}')
                self.timer.relative_click(0.38, 0.565)
                destroy_ship(self.timer)
                return self.run(retry_times + 1)
            return ret
        elif ret == ConditionFlag.FIGHT_END:
            self.timer.set_page(self.info.end_page)
            return ret
        elif ret == ConditionFlag.BATTLE_TIMES_EXCEED or ret == ConditionFlag.SKIP_FIGHT:
            return ret
        else:
            self.logger.error('无法进入战斗, 原因未知! 屏幕状态已记录')
            self.timer.log_screen()
            raise BaseException(str(time.time()) + 'enter fight error')

        # 战斗中逻辑
        return self.fight()

    def run_for_times_condition(
        self,
        times,
        last_point,
        result='S',
        insist_time=900,
    ) -> ConditionFlag | bool:
        """有战果要求的多次运行, 使用前务必检查参数是否有误, 防止死循环

        Args:
            times: 次数

            last_point: 最后一个点

            result: 战果要求

            insist_time: 如果大于这个时间工作量未减少则退出工作

        Returns:
            ConditionFlag
            False: 不满足预设条件, 此次战斗不计入次数
        """
        if not isinstance(result, str) or not isinstance(last_point, str):
            raise TypeError(
                f'last_point, result must be str,but is {type(last_point)}, {type(result)}',
            )
        if result not in ['S', 'A', 'B', 'C', 'D', 'SS']:
            raise ValueError(
                f"result value {result} is illegal, it should be 'A','B','C','D','S' or 'SS'",
            )
        if len(last_point) != 1 or ord(last_point) > ord('Z') or ord(last_point) < ord('A'):
            raise ValueError("last_point should be a uppercase within 'A' to 'Z'")
        import time

        result_list = ['SS', 'S', 'A', 'B', 'C', 'D']
        start_time = time.time()
        while times:
            ret = self.run()
            if ret == ConditionFlag.DOCK_FULL:
                self.timer.logger.error('船坞已满, 无法继续')
                return ret

            self.logger.info('战斗信息:\n' + str(self.info.fight_history))
            fight_results = sorted(self.info.fight_history.get_fight_results().items())
            # 根据情况截取战果，并在result_list查找索引
            if len(fight_results):
                if str(fight_results[-1][1])[-2].isalpha():
                    fight_result_index = result_list.index(
                        str(fight_results[-1][1])[-2:],
                    )
                else:
                    fight_result_index = result_list.index(
                        str(fight_results[-1][1])[-1],
                    )

            finish = (
                len(fight_results)
                and fight_results[-1][0] == last_point
                and fight_result_index <= result_list.index(result)
            )
            if not finish:
                self.timer.logger.info(
                    f'不满足预设条件, 此次战斗不计入次数, 剩余战斗次数:{times}',
                )
                if time.time() - start_time > insist_time:
                    return False
            else:
                start_time, times = time.time(), times - 1
                self.timer.logger.info(
                    f'完成了一次满足预设条件的战斗, 剩余战斗次数:{times}',
                )
        return ConditionFlag.OPERATION_SUCCESS

    def update_state(self, *args, **kwargs):
        try:
            self.info.update_state()
            state = self.info.state
            self.timer.keep_try_update_fight = 0
        except ImageNotFoundErr as _:
            # 处理点击延迟或者网络波动导致的匹配失败
            if (
                hasattr(self.timer, 'keep_try_update_fight')
                and self.timer.keep_try_update_fight > 3
            ):
                return ConditionFlag.SL
            if hasattr(self.timer, 'keep_try_update_fight'):
                self.timer.keep_try_update_fight += 1
            else:
                self.timer.keep_try_update_fight = 1
            self.logger.warning('Image Match Failed, Trying to Process')
            if self.timer.is_other_device_login():
                self.timer.process_other_device_login()  # TODO: 处理其他设备登录
            if self.timer.is_bad_network(timeout=5):
                self.timer.process_bad_network(extra_info='update_state', timeout=5)
            self._make_decision(skip_update=True)
            # if self.info.last_state == "spot_enemy_success":
            #     if self.timer.image_exist(IMG.fight_image[2]):
            #         self.timer.click(900, 500)
            # if self.info.last_state in ["proceed", "night"] and self.timer.image_exist(
            #     IMG.fight_image[5:7]
            # ):
            #     if self.info.last_action == "yes":
            #         self.timer.click(325, 350, times=1)
            #     else:
            #         self.timer.click(615, 350, times=1)

            if 'try_times' not in kwargs:
                return self.update_state(try_times=1)
            time.sleep(10 * 2.5 ** kwargs['try_times'])
            return self.update_state(try_times=kwargs['try_times'] + 1)
        return state

    def _enter_fight(self) -> ConditionFlag:
        """进入战斗前的操作"""

    def _make_decision(self, *args, **kwargs) -> str:
        """决策函数"""

    # =============== 战斗中通用的操作 ===============
    def _sl(self):
        self.timer.logger.debug('正在执行SL操作')
        # 重置地图节点信息
        self.timer.reset_chapter_map()

        self.timer.restart()
        self.timer.go_main_page()
        self.timer.set_page('main_page')


class DecisionBlock:
    """地图上一个节点的决策模块"""

    def __init__(self, timer: Timer, args) -> None:
        self.timer = timer
        self.logger = timer.logger
        self.config = NodeConfig.from_dict(args)

        # 用于根据规则设置阵型
        self.set_formation_by_rule = False
        self.formation_by_rule = Formation.double_column

    def _check_rules(self, enemies: dict) -> SearchEnemyAction | Formation:
        for rule in self.config.enemy_rules:
            condition, act = rule
            eval_condition = ''
            last = 0
            for i, ch in enumerate(condition):
                if ord(ch) > ord('Z') or ord(ch) < ord('A'):
                    if last != i:
                        if condition[last:i] in ALL_SHIP_TYPES:
                            eval_condition += str(enemies.get(condition[last:i], 0))
                        else:
                            eval_condition += condition[last:i]
                    eval_condition += ch
                    last = i + 1

            condition_result = eval(eval_condition)
            if self.timer.config.show_enemy_rules:
                act_info = f'判断敌舰规则: {condition}, 结果: {condition_result}'
                if condition_result:
                    act_info += ', 执行: '
                    act_info += act if isinstance(act, str) else f'选择阵型: {act}'
                else:
                    act_info += ', 不执行特殊操作进入战斗'
                self.logger.info(act_info)
            if condition_result:
                if isinstance(act, str):
                    return SearchEnemyAction(act)
                return Formation(act)
        return SearchEnemyAction.no_action

    def _check_formation_rules(self, formation: str) -> SearchEnemyAction | Formation:
        for rule in self.config.enemy_formation_rules:
            condition, act = rule
            condition_result = condition == formation
            act_info = f'判断敌舰阵容规则: {condition}, 结果: {condition_result}'
            if condition_result:
                act_info += f', 执行: {act}'
                if isinstance(act, str):
                    return SearchEnemyAction(act)
                return Formation(act)
            act_info += ', 不执行特殊操作进入战斗'
        return SearchEnemyAction.no_action

    def make_decision(self, state, last_state, last_action, info: FightInfo):
        # destroy_ship skip: extract-method
        """单个节点的决策"""
        enemies = info.enemies
        if state in ['fight_period', 'night_fight_period']:
            if self.config.SL_when_enter_fight:
                info.fight_history.add_event(
                    '进入战斗',
                    {
                        'position': (
                            info.node
                            if 'node' in info.__dict__
                            else f'此类战斗({type(info)})不包含节点信息'
                        ),
                    },
                    'SL',
                )
                return None, ConditionFlag.SL
            return None, ConditionFlag.FIGHT_CONTINUE

        if state == 'spot_enemy_success':
            retreat = False
            can_detour = self.timer.image_exist(
                IMG.fight_image[13],
            )  # 判断该点是否可以迂回
            detour = can_detour and self.config.detour  # 由 Node 指定是否要迂回

            # 功能, 根据敌方阵容进行选择
            act = self._check_formation_rules(info.enemy_formation)
            if act == SearchEnemyAction.no_action:
                act = self._check_rules(enemies=enemies)

            if act == SearchEnemyAction.retreat:
                retreat = True
            elif act == SearchEnemyAction.detour:
                try:
                    assert can_detour, '该点无法迂回, 但是规则中指定了迂回'
                except AssertionError:
                    raise ValueError('该点无法迂回, 但在规则中指定了迂回')
                detour = True
            elif isinstance(act, Formation):
                self.set_formation_by_rule = True
                self.formation_by_rule = act

            if retreat:
                self.timer.click(677, 492, delay=0.2)
                info.fight_history.add_event(
                    '索敌成功',
                    {
                        'position': (
                            info.node
                            if 'node' in info.__dict__
                            else f'此类战斗({type(info)})不包含节点信息'
                        ),
                    },
                    '撤退',
                )
                return 'retreat', ConditionFlag.FIGHT_END
            if detour:
                image_detour = IMG.fight_image[13]
                if self.timer.click_image(image=image_detour, timeout=2.5):
                    self.timer.logger.info('成功执行迂回操作')
                else:
                    self.timer.logger.error('未找到迂回按钮')
                    self.timer.log_screen(True)
                    raise ImageNotFoundErr("can't found image")

                # self.timer.click(540, 500, delay=0.2)
                info.fight_history.add_event(
                    '索敌成功',
                    {
                        'position': (
                            info.node
                            if 'node' in info.__dict__
                            else f'此类战斗({type(info)})不包含节点信息'
                        ),
                    },
                    '迂回',
                )
                return 'detour', ConditionFlag.FIGHT_CONTINUE

            info.fight_history.add_event(
                '索敌成功',
                {
                    'position': (
                        info.node
                        if 'node' in info.__dict__
                        else f'此类战斗({type(info)})不包含节点信息'
                    ),
                },
                '战斗',
            )
            if self.config.long_missile_support:
                image_missile_support = IMG.fight_image[17]
                if self.timer.click_image(image=image_missile_support, timeout=2.5):
                    self.timer.logger.info('成功开启远程导弹支援')
                else:
                    self.timer.logger.error('未找到远程支援按钮')
                    raise ImageNotFoundErr("can't found image of long_missile_support")
            time.sleep(0.5)
            self.timer.click(855, 501, delay=0.2)  # 0.891 0.928
            # self.timer.click(380, 520, times=2, delay=0.2) # TODO: 跳过可能的开幕支援动画，实现有问题
            return 'fight', ConditionFlag.FIGHT_CONTINUE
        if state == 'formation':
            spot_enemy = last_state == 'spot_enemy_success'
            value = self.config.formation
            if spot_enemy:
                if self.config.SL_when_detour_fails and last_action == 'detour':
                    info.fight_history.add_event(
                        '迂回',
                        {
                            'position': (
                                info.node
                                if 'node' in info.__dict__
                                else f'此类战斗({type(info)})不包含节点信息'
                            ),
                        },
                        result='失败',
                    )
                    info.fight_history.add_event(
                        '阵型选择',
                        {
                            'enemies': enemies,
                            'position': (
                                info.node
                                if 'node' in info.__dict__
                                else f'此类战斗({type(info)})不包含节点信息'
                            ),
                        },
                        action='SL',
                    )
                    return None, ConditionFlag.SL

                if self.set_formation_by_rule:
                    self.logger.debug('set formation by rule:', self.formation_by_rule)
                    value = self.formation_by_rule
                    self.set_formation_by_rule = False
            else:
                if self.config.SL_when_spot_enemy_fails:
                    info.fight_history.add_event(
                        '阵型选择',
                        {
                            'enemies': '索敌失败',
                            'position': (
                                info.node
                                if 'node' in info.__dict__
                                else f'此类战斗({type(info)})不包含节点信息'
                            ),
                        },
                        action='SL',
                    )
                    return None, ConditionFlag.SL
                if self.config.formation_when_spot_enemy_fails:
                    value = self.config.formation_when_spot_enemy_fails
            info.fight_history.add_event(
                '阵型选择',
                {
                    'enemies': (enemies if last_state == 'spot_enemy_success' else '索敌失败'),
                    'position': (
                        info.node
                        if 'node' in info.__dict__
                        else f'此类战斗({type(info)})不包含节点信息'
                    ),
                },
                action=value,
            )
            self.timer.relative_click(*value.relative_position, delay=2)
            return value, ConditionFlag.FIGHT_CONTINUE
        if state == 'night':
            is_night = self.config.night
            info.fight_history.add_event(
                '是否夜战',
                {
                    'position': (
                        info.node
                        if 'node' in info.__dict__
                        else f'此类战斗({type(info)})不包含节点信息'
                    ),
                },
                action='追击' if is_night else '撤退',
            )

            match_night(self.timer, is_night)
            if is_night:
                # self.timer.click(325, 350)
                return 'yes', ConditionFlag.FIGHT_CONTINUE
            # self.timer.click(615, 350)
            return 'no', ConditionFlag.FIGHT_CONTINUE

        if state == 'result':
            # time.sleep(1.5)
            # self.timer.click(900, 500, times=2, delay=0.2)
            click_result(self.timer)
            return None, ConditionFlag.FIGHT_CONTINUE
        if state == 'get_ship':
            get_ship(self.timer)
            return None, ConditionFlag.FIGHT_CONTINUE
        self.logger.error('Unknown State')
        raise BaseException



import time
from typing import Any

from autowsgr.utils.logger import Logger


class Ship:
    def __init__(self, name) -> None:
        self.level = 1
        self.name = name  # 舰船名: 舰船的唯一标识
        self.type = 'DD'  # 舰船类型, 暂时没啥用
        self._statu = 0  # 数字: 0 绿血, 1 黄血, 2 红血, 3 修理中.
        self.repair_end_time = 0
        self.repair_start_time = 0
        self.waiting_repair = 0  # 是否正在修理队列中
        self.equipments = None  # 装备状态, 暂时不用

    def __str__(self) -> str:
        table = ['绿血', '中破', '大破', '修理中']
        return f'舰船名:{self.name}\n舰船状态:{table[self.status]}\n舰船等级:{self.level}\n\n'

    @property
    def status(self):
        if self.is_repairing():
            return 3
        return self._statu

    @status.setter
    def status(self, new_statu):
        self._statu = new_statu

    def set_repair(self, time_cost):
        """设置舰船正在修理中
        Args:
            time_cost (str): 识别出的时间字符串
        """
        self.waiting_repair = False
        self._statu = 0
        self.repair_start_time = time.time()
        self.repair_end_time = self.repair_start_time + time_cost

    def is_repairing(self):
        return time.time() < self.repair_end_time


class WorkShop:
    def __init__(self) -> None:
        self.available_time = None

    def _time_to_seconds(self, time_str):
        parts = time_str.split(':')
        return int(parts[0]) * 3600 + int(parts[1]) * 60 + int(parts[2])

    def update_available_time(self, available_time):
        self.available_time = available_time

    def get_waiting_time(self):
        # 如果不确定, 返回 None
        if self.available_time is None:
            return None

        waiting_time = 86400
        for bath in self.available_time:
            if time.time() > bath:
                return 0
            waiting_time = min(waiting_time, bath - time.time() + 0.1)
        return waiting_time

    def is_available(self):
        return self.get_waiting_time() == 0

    def add_work(self, time_cost: str) -> tuple[bool, float]:
        """增加一项工作

        Args:
            time_cost (str): _description_

        Returns:
            (bool, float): bool 值表示是否成功添加, float 表示结束时间

        """
        if self.is_available() and self.available_time is not None:
            for id, bath in enumerate(self.available_time):
                if time.time() > bath:
                    end_time = time.time() + self._time_to_seconds(time_cost)
                    self.available_time[id] = end_time
                    return (True, end_time)
        return (False, 0)


class BathRoom(WorkShop):
    def __init__(self) -> None:
        super().__init__()

    def add_repair(self, time_cost: str) -> None:
        super().add_work(time_cost)


class Factory(WorkShop):
    def __init__(self) -> None:
        self.capacity = None
        self.waiting_destory = False

    def update_capacity(self, capacity, occupation, blueprint: int | None = None):
        """更新仓库容量状态

        Args:
            capacity : 总容量
            occupation : 总占用量
        """
        self.capacity = int(capacity)
        self.occupation = int(occupation)
        if blueprint is not None:
            self.blueprint = blueprint

    @property
    def full(self):
        if self.capacity is not None:
            return self.occupation >= self.capacity
        return False


class Port:
    factory: Any

    def __init__(self, logger: Logger) -> None:
        self.logger = logger
        self.oil = 0
        self.ammo = 0
        self.steel = 0
        self.aluminum = 0
        self.diamond = 0
        self.quick_build = 0
        self.bathroom = BathRoom()
        self.ship_factory = Factory()
        self.ships: list[Ship] = []
        self.fleet = [[]] * 5
        self.map: int | None = 0
        self.chapter: int | None = 0

    def have_ship(self, name):
        return any(name == ship.name for ship in self.ships)

    def register_ship(self, name):
        if not self.have_ship(name):
            ship = Ship(name)
            self.ships.append(ship)
            self.logger.info(f'舰船 {name} 已注册')
            return ship
        return None

    def get_ship_by_name(self, name) -> Ship | None:
        if self.have_ship(name):
            for ship in self.ships:
                if ship.name == name:
                    return ship
        self.show_fleet()
        self.logger.info(f'需要查找的舰船 {name} 未找到')
        return None

    def show_fleet(self):
        self.logger.info('当前已经注册的舰船如下:')
        for ship in self.ships:
            print(ship)



import datetime
from typing import Any

from autowsgr.constants.image_templates import IMG
from autowsgr.game.game_operation import get_ship
from autowsgr.timer import Timer
from autowsgr.utils.api_image import absolute_to_relative, crop_image, match_nearest_index
from autowsgr.utils.time import get_eta, str2time


# 开始建造，建造完成，快速建造 三个按钮的中心位置
BUILD_POSITIONS = {
    'ship': [
        (0.823, 0.312),
        (0.823, 0.508),
        (0.823, 0.701),
        (0.823, 0.898),
    ],
}
# x轴边缘到中心距离，y轴边缘到中心距离，y轴相对于BUILD上移
ETA_DELTA = [0.066, 0.025, 0.086]
# 剩余时间区域
ETA_AREAS = {
    'ship': [
        (
            (pos[0] - ETA_DELTA[0], pos[1] + ETA_DELTA[1] - ETA_DELTA[2]),
            (pos[0] + ETA_DELTA[0], pos[1] - ETA_DELTA[1] - ETA_DELTA[2]),
        )
        for pos in BUILD_POSITIONS['ship']
    ],
}
# 装备页面，右侧整体下移
EQUIPMENT_DELTA = 0.02
BUILD_POSITIONS['equipment'] = [
    (pos[0], pos[1] + EQUIPMENT_DELTA) for pos in BUILD_POSITIONS['ship']
]
ETA_AREAS['equipment'] = [
    (
        (area[0][0], area[0][1] + EQUIPMENT_DELTA),
        (area[1][0], area[1][1] + EQUIPMENT_DELTA),
    )
    for area in ETA_AREAS['ship']
]
# 四个资源的左下角位置
RESOURCE_POSITIONS = [(0.2, 0.455), (0.59, 0.455), (0.2, 0.855), (0.59, 0.855)]
# 四个资源的区域
RESOURCE_AREAS = [(pos, (pos[0] + 0.16, pos[1] - 0.1)) for pos in RESOURCE_POSITIONS]
# 操作资源的位置（第一位）
RESOURCE_OPERATE_POSITIONS = [
    [(pos[0] + 0.022 + i * 0.057, pos[1] - 0.046) for i in range(3)] for pos in RESOURCE_POSITIONS
]
# 滑动的距离
RESOURCE_OPERATE_DELTA = 0.09


class BuildManager:
    # TODO: 获取建造舰船名称; 正确处理异常情况; 使用内部记忆代替查询
    def __init__(self, timer: Timer) -> None:
        self.timer = timer
        # -1代表空位. None代表未开通
        self.slot_eta: dict[str, list[Any]] = {
            'ship': [None] * 4,
            'equipment': [None] * 4,
        }
        self.initialized = False

    def init(self):
        self.update_slot_eta('ship')
        self.update_slot_eta('equipment')
        self.initialized = True

    def update_slot_eta(self, type='ship'):
        """更新建造队列的剩余时间"""
        # 进入页面
        if type == 'equipment':
            self.timer.goto_game_page('develop_page')
        if type == 'ship':
            self.timer.goto_game_page('build_page')
        # 截图检测
        screen = self.timer.get_screen(self.timer.resolution, need_screen_shot=True)
        for build_slot in range(4):
            ocr_result = self.timer.recognize(
                crop_image(screen, *ETA_AREAS[type][build_slot]),
                allow_nan=True,
            )
            if not ocr_result:
                self.slot_eta[type][build_slot] = None
            elif '完成' in ocr_result[1] or '开始' in ocr_result[1]:
                self.slot_eta[type][build_slot] = -1
            elif ':' in ocr_result[1]:
                self.slot_eta[type][build_slot] = get_eta(str2time(ocr_result[1]))

    def get_timedelta(self, type='ship'):
        """获取建造队列的最小剩余时间
        Args:
            type (str): "ship"/"equipment"
        Returns:
            datetime.timedelta: 剩余时间
        """
        valid_times = [eta for eta in self.slot_eta[type] if eta not in (-1, None)]
        self.timer.logger.debug(valid_times)
        return min(valid_times) - datetime.datetime.now() if valid_times else datetime.timedelta()

    def get_min_eta(self, type='ship'):
        """获取建造队列的最小结束时间
        Args:
            type (str): "ship"/"equipment"
        Returns:
            datetime.datetime: 结束时间
        """
        return min(self.slot_eta[type])

    def has_empty_slot(self, type='ship') -> bool:
        """检查是否有空位
        Args:
            type (str): "ship"/"equipment"
        Returns:
            bool: 是否有空位
        """
        if not self.initialized:
            self.init()
        # 检查是否包含 -1 或当前时间是否大于任何非 None 的时间
        return -1 in self.slot_eta[type] or any(
            datetime.datetime.now() > eta for eta in self.slot_eta[type] if eta is not None
        )

    def get_build(self, type='ship', allow_fast_build=False) -> bool:
        """获取已经建造好的舰船或装备
        Args:
            type (str): "ship"/"equipment"
            allow_fast_build: 是否允许快速建造
        Returns:
            bool: 是否获取成功
        """
        # 进入页面
        if type == 'equipment':
            self.timer.goto_game_page('develop_page')
        if type == 'ship':
            self.timer.goto_game_page('build_page')

        # 快速建造
        if allow_fast_build:
            while self.timer.image_exist(IMG.build_image[type].fast):
                self.timer.click_image(IMG.build_image[type].fast)
                self.timer.confirm_operation(must_confirm=True, timeout=3)

        # 收完成
        while self.timer.image_exist(IMG.build_image[type].complete):
            try:
                pos = self.timer.click_image(
                    IMG.build_image[type].complete,
                    timeout=3,
                    must_click=True,
                )
                if self.timer.image_exist(IMG.build_image[type].full_depot):
                    self.timer.logger.warning(f'{type} 仓库已满')
                    self.timer.go_main_page()
                    return False
            except Exception as e:
                self.timer.logger.warning(f'收取 {type} 失败: {e}')
                return False

            try:
                ship_name, ship_type = get_ship(self.timer)
                self.timer.logger.info(f'获取 {type}: {ship_name} {ship_type}')
            except Exception as e:
                self.timer.logger.warning(f'识别获取{type}内容失败: {e}')

            slot = match_nearest_index(
                absolute_to_relative(pos, self.timer.resolution),
                BUILD_POSITIONS[type],
            )
            self.slot_eta[type][slot] = -1
        return True

    def build(self, type='ship', resources=None, allow_fast_build=False):
        """建造操作
        Args:
            type (str): "ship"/"equipment"
            resources: 一个列表, 表示油弹钢铝四项资源. Defaults to None.
            allow_fast_build (bool, optional): 如果队列已满, 是否立刻结束一个以开始建造. Defaults to False.
        """

        def choose_build_resources():
            """移动资源
            Args:
                resource_id (int): 资源编号
                dst (int): 目标值
            """
            # 无需设置资源的情况
            if not resources:
                return

            def value_to_digits(value):
                # 拆分为3位数
                return [value // 100, value // 10 % 10, value % 10]

            def detect_build_resources(resource_id) -> list:
                """检查四项资源余量
                Args:
                    resource_id (int): 具体哪一项资源 [0,3]
                Returns:
                    list: 拆分为3位数的资源
                """
                screen = self.timer.get_screen()
                recognize_result = self.timer.recognize_number(
                    crop_image(screen, *RESOURCE_AREAS[resource_id]),
                    rgb_select=(255, 155, 81),
                )
                assert recognize_result is not None
                value = recognize_result[1]
                return value_to_digits(value)

            resource_digits = [value_to_digits(res) for res in resources]
            for resource_id, dst in enumerate(resource_digits):
                src = detect_build_resources(resource_id)
                for digit in range(3):
                    while src[digit] != dst[digit]:
                        print(f'资源 {resource_id} 目前 {src} 目标 {dst}')
                        way = -1 if src[digit] < dst[digit] else 1
                        self.timer.relative_swipe(
                            *RESOURCE_OPERATE_POSITIONS[resource_id][digit],
                            RESOURCE_OPERATE_POSITIONS[resource_id][digit][0],
                            RESOURCE_OPERATE_POSITIONS[resource_id][digit][1]
                            + way * RESOURCE_OPERATE_DELTA,
                            duration=0.25,
                        )
                        src = detect_build_resources(resource_id)

        # 检查资源有效性
        if resources:
            minv = 30 if type == 'ship' else 10
            maxv = 999
            if min(resources) < minv or max(resources) > maxv:
                self.timer.logger.warning(
                    f'用于 {type} 的资源 {resources} 越界, 已自动取消操作',
                )
                return False

        # 收完成，检查空队列
        if not self.get_build(type, allow_fast_build):
            return False
        if not self.slot_eta[type].count(-1):
            self.timer.logger.warning(f'{type} 建造队列已满')
            return False

        # 点击建造
        self.timer.click_image(IMG.build_image[type].start)
        # 选择资源，开始建造
        self.timer.wait_image(IMG.build_image.resource)
        choose_build_resources()
        self.timer.relative_click(0.89, 0.89)
        # 更新建造时间
        self.update_slot_eta(type)

        return True



import copy

from autowsgr.configs import ExerciseConfig
from autowsgr.constants.image_templates import IMG
from autowsgr.fight.common import DecisionBlock, FightInfo, FightPlan, start_march
from autowsgr.game.game_operation import detect_ship_stats, move_team, quick_repair
from autowsgr.game.get_game_info import get_enemy_condition, get_exercise_stats
from autowsgr.timer import Timer
from autowsgr.types import ConditionFlag, Formation, SearchEnemyAction
from autowsgr.utils.io import recursive_dict_update, yaml_to_dict


"""
演习决策模块
"""


class ExerciseDecisionBlock(DecisionBlock):
    def __init__(
        self,
        timer: Timer,
        node_args: dict,
        max_refresh_times: int,
        fleet_id: int,
    ) -> None:
        super().__init__(timer, node_args)
        self.max_refresh_times = max_refresh_times
        self.fleet_id = fleet_id

    def make_decision(self, state, last_state, last_action, info: FightInfo):
        if state == 'rival_info':
            max_times = self.max_refresh_times
            self.formation_chosen = self.config.formation
            while max_times >= 0:
                info.enemies = get_enemy_condition(self.timer)
                act = self._check_rules(info.enemies)
                if act == SearchEnemyAction.refresh:
                    if max_times > 0:
                        max_times -= 1
                        self.timer.click(665, 400, delay=0.75)
                    else:
                        break
                elif isinstance(act, Formation):
                    self.formation_chosen = act
                elif act is SearchEnemyAction.no_action:
                    break

            self.timer.click(804, 390, delay=0)
            return 'fight', ConditionFlag.FIGHT_CONTINUE

        if state == 'fight_prepare_page':
            move_team(self.timer, self.fleet_id)
            info.ship_stats = detect_ship_stats(self.timer)
            quick_repair(self.timer, ship_stats=info.ship_stats)
            if start_march(self.timer) != ConditionFlag.OPERATION_SUCCESS:
                return self.make_decision(state, last_state, last_action, info)
            return None, ConditionFlag.FIGHT_CONTINUE

        if state == 'spot_enemy_success':
            self.timer.click(900, 500, delay=0)
            return None, ConditionFlag.FIGHT_CONTINUE

        if state == 'formation':
            self.timer.click(573, self.formation_chosen * 100 - 20, delay=2)
            return None, ConditionFlag.FIGHT_CONTINUE

        return super().make_decision(state, last_state, last_action, info)


class NormalExerciseInfo(FightInfo):
    """存储战斗中需要用到的所有状态信息"""

    def __init__(self, timer: Timer) -> None:
        super().__init__(timer)
        self.end_page = 'exercise_page'
        self.successor_states = {
            'exercise_page': ['rival_info'],
            'rival_info': {
                'fight': ['fight_prepare_page'],
                'discard': ['exercise_page'],
            },
            'fight_prepare_page': ['spot_enemy_success', 'formation', 'fight_period'],
            'spot_enemy_success': ['formation', 'fight_period'],
            'formation': ['fight_period'],
            'fight_period': ['night', 'result'],
            'night': {
                'yes': ['night_fight_period'],
                'no': [['result', 7]],
            },
            'night_fight_period': ['result'],
            'result': ['exercise_page'],  # 两页战果
        }

        self.state2image = {
            'exercise_page': [IMG.identify_images['exercise_page'], 7.5],
            'rival_info': [IMG.exercise_image['rival_info'], 7.5],
            'fight_prepare_page': [IMG.identify_images['fight_prepare_page'], 7.5],
            'spot_enemy_success': [IMG.fight_image[2], 15],
            'formation': [IMG.fight_image[1], 15],
            'fight_period': [IMG.symbol_image[4], 10],
            'night': [IMG.fight_image[6], 0.85, 180],
            'night_fight_period': [IMG.symbol_image[4], 10],
            'result': [IMG.fight_image[3], 90],
        }

        self.after_match_delay = {
            'night': 1.75,
        }

    def reset(self):
        self.fight_history.reset()
        self.state = 'rival_info'  # 初始状态等同于 "rival_info" 选择 'discard'
        self.last_action = 'discard'

    def _before_match(self):
        # 点击加速
        if self.state in ['fight_prepare_page']:
            self.timer.click(
                380,
                520,
                delay=0,
                enable_subprocess=True,
            )

        self.timer.update_screen()

    @property
    def node(self):
        return self.timer.ship_point


class NormalExercisePlan(FightPlan):
    """
    常规战斗的决策模块

    Args:
        plan_path: 以 PLAN_ROOT/exercise 为根的相对路径

        fleet_id: 指定舰队编号, 如果为 None 则使用计划中的参数

    """

    def __init__(self, timer: Timer, plan_path: str, fleet_id: int | None) -> None:
        super().__init__(timer)

        # 加载计划配置
        plan_args = yaml_to_dict(self.timer.plan_tree['exercise'][plan_path])
        if fleet_id is not None:
            plan_args['fleet_id'] = fleet_id  # 舰队编号
        assert 'fleet_id' in plan_args, '未指定作战舰队'
        self.config = ExerciseConfig.from_dict(plan_args)

        self.nodes: dict[str, ExerciseDecisionBlock] = {}
        for node_name in self.config.selected_nodes:
            node_args = copy.deepcopy(plan_args.get('node_defaults', {}))
            if node_name in plan_args['node_args']:
                node_args = recursive_dict_update(
                    node_args,
                    plan_args['node_args'][node_name],
                )
            self.nodes[node_name] = ExerciseDecisionBlock(
                timer,
                node_args,
                self.config.max_refresh_times,
                self.config.fleet_id,
            )

        # 构建信息存储结构
        self.info = NormalExerciseInfo(self.timer)

    def _enter_fight(self) -> str:
        """
        从任意界面进入战斗.

        Returns:
            ConditionFlag
        """
        self.timer.goto_game_page('exercise_page')
        self._exercise_times = self.config.exercise_times
        self.exercise_stats = [None, None]
        return ConditionFlag.OPERATION_SUCCESS

    def _make_decision(self, *args, **kwargs):
        state = self.update_state() if 'skip_update' not in kwargs else self.info.state
        if state == ConditionFlag.SL:
            return ConditionFlag.SL

        # 进行MapLevel的决策
        if state == 'exercise_page':
            self.exercise_stats = get_exercise_stats(self.timer, self.exercise_stats[1])
            if self._exercise_times > 0 and any(self.exercise_stats[2:]):
                pos = self.exercise_stats[2:].index(True)
                self.rival = 'player'
                self.timer.click(770, (pos + 1) * 110 - 10)
                return ConditionFlag.FIGHT_CONTINUE
            if 'robot' in self.config.selected_nodes and self.exercise_stats[1]:
                self.timer.swipe(800, 200, 800, 400)  # 上滑
                self.timer.click(770, 100)
                self.rival = 'robot'
                self.exercise_stats[1] = False
                return ConditionFlag.FIGHT_CONTINUE

            return ConditionFlag.FIGHT_END

        # 进行通用NodeLevel决策
        action, fight_stage = self.nodes[self.rival].make_decision(
            state,
            self.info.last_state,
            self.info.last_action,
            self.info,
        )
        self.info.last_action = action
        return fight_stage



import re

import requests


# os.environ["https_proxy"] = "127.0.0.1:7890"
# os.environ["http_proxy"] = "127.0.0.1:7890"

UPDATE = True
URL = 'https://www.zjsnrwiki.com/wiki/%E8%88%B0%E5%A8%98%E5%9B%BE%E9%89%B4#searchInput'
HTML_PATH = './ship_name.html'
YAML_PATH = './ship_name_example.yaml'


def get_source():
    if UPDATE:
        response = requests.get(URL, timeout=5)
        html = response.text
        with open(HTML_PATH, mode='w+', encoding='utf-8') as f:
            f.write(html)
    else:
        with open(HTML_PATH, encoding='utf-8') as f:
            html = f.read()
    return html


def extract(str):
    re_rk_wsrwiki = r'<td width="162px"><center><b>(.*?)</b></center></td>'
    re_name_wsrwiki = (
        r'<td width="162px" height="56px"><center><b><a [^>]*>(.*?)</a></b></center></td>'
    )
    res = ''
    rks = re.findall(re_rk_wsrwiki, str)
    names = re.findall(re_name_wsrwiki, str)
    # print(rks)
    # print(names)
    print(len(rks), len(names))

    for rk, name in zip(rks, names, strict=False):
        rk = rk[3:].strip()  # 添加.strip()以去除可能的空格和换行符
        # 获取舰船名字
        name = name.strip()
        _name = name[: name.find('(')] if name.find('(') != -1 else name
        res += f'No.{rk}: # {name}\n'
        res += f'  - "{_name}"\n'

    res += """Other: # 战例
  - 肌肉记忆
  - 长跑训练
  - 航空训练
  - 训练的结晶
  - 黑科技
  - 防空伞
  - 守护之盾
  - 抱头蹲防
  - 关键一击
  - 久远的加护"""
    return res


with open(YAML_PATH, mode='w+', encoding='utf-8') as f:
    f.write(extract(get_source()))



from urllib import parse, request

import yaml


def load_config():
    with open('miao_alert.yaml', encoding='utf-8') as file:
        return yaml.safe_load(file)


def miao_alert(index):
    try:
        config = load_config()
        miao_code = config['miao_code']
        texts = config['texts']  # 从 YAML 文件中读取 texts 列表

        # 检查索引是否在列表范围内
        if 0 <= index < len(texts):
            text = texts[index]
            request.urlopen(
                'https://miaotixing.com/trigger?'
                + parse.urlencode({'id': miao_code, 'text': text}),
            )
        else:
            print('指定的索引超出了 texts 列表的范围。')
    except Exception as e:
        print(f'执行 miao_alert 函数时发生错误：{e}')
        # 可以选择记录日志或执行其他错误处理操作


if __name__ == '__main__':
    # 调用函数时指定索引，例如使用列表中的第二个 text
    miao_alert(1)



import os
import pathlib
import sys


sys.path.append(os.path.dirname(os.path.dirname(__file__)))

import cv2
import keyboard
from numpy.typing import NDArray

from autowsgr.scripts.main import start_script
from autowsgr.timer import Timer
from autowsgr.utils.io import dict_to_yaml, listdir


# en_reader = easyocr.Reader(['en'], gpu=False)
timer = None
point = 'A'
screen_shot_count = 0


def log_image(event: keyboard.KeyboardEvent):
    global screen_shot_count
    assert isinstance(timer, Timer)
    if event.event_type != 'down' or event.name != 'P':
        return
    print('Screen Processing:', screen_shot_count)
    screen_shot_count += 1
    timer.update_screen()
    timer.log_screen()


def set_points(windowname, img: NDArray):
    """
    输入图片，打开该图片进行标记点，返回的是标记的几个点的字符串和相对坐标
    """
    global point
    point = 'A'
    print('(提示：单击需要标记的坐标，Enter确定，Esc跳过，其它重试。)')
    points = {}
    relative_points = {}

    def on_mouse(event, x, y, flags, param):
        global point
        if event == cv2.EVENT_LBUTTONDOWN:
            cv2.circle(temp_img, (x, y), 10, (102, 217, 239), -1)
            points[point] = (x, y)
            relative_points[point] = (x / img.shape[1], y / img.shape[0])
            point = chr(ord(point) + 1)
            cv2.imshow(windowname, temp_img)

    temp_img = img.copy()
    cv2.namedWindow(windowname)
    cv2.imshow(windowname, temp_img)
    cv2.setMouseCallback(windowname, on_mouse)
    key = cv2.waitKey(0)
    if key == 13:  # Enter
        print('坐标为：', points)
        print('相对坐标为：', relative_points)
        del temp_img
        cv2.destroyAllWindows()
        str(points)
    elif key == 27:  # ESC
        print('跳过该张图片')
        del temp_img
        cv2.destroyAllWindows()
    else:
        print('重试!')
        set_points(windowname, img)

    print(points)
    print(relative_points)
    return points, relative_points


def get_image() -> None:
    global timer
    timer = start_script('../examples/user_settings.yaml')
    import time

    keyboard.hook(log_image)
    time.sleep(1000)


def make_map(image_path: str, dict_dir: str) -> None:
    """根据图像目录下的所有图片文件,打开后顺次点击ABCD,生成对应文件名的地图文件

    Args:
        image_path (_type_): _description_
        dict_dir (_type_): _description_
    """
    files = listdir(image_path)
    for file in files:

        f = pathlib.Path(file)
        if f.suffix != '.PNG':
            continue
        name = f.stem
        dict_value, relative_value = set_points(name, cv2.imread(file))
        dict_to_yaml(dict_value, os.path.join(dict_dir, 'E-' + name[1:] + '.yaml'))
        dict_to_yaml(dict_value, os.path.join(dict_dir, 'H-' + name[1:] + '.yaml'))
        dict_to_yaml(
            relative_value,
            os.path.join(dict_dir, 'E-' + name[1:] + '_relative.yaml'),
        )


if __name__ == '__main__':
    print(
        """Input operation type:
          1: log image when 'P' pressed.
          2: make map .yaml files""",
    )
    oper = input().split()[0]
    if oper == '1':
        get_image()
    elif oper == '2':
        print('Enter image_path:')
        image_path = input()
        print('Enter dict_path')
        dict_path = input()
        make_map(image_path, dict_path)



import copy
import datetime
import threading as th
import time
from collections.abc import Iterable
from concurrent.futures import ProcessPoolExecutor
from typing import Any

import cv2
from airtest.core.android import Android
from numpy import uint8
from numpy.typing import NDArray

from autowsgr.configs import UserConfig
from autowsgr.constants.custom_exceptions import CriticalErr, ImageNotFoundErr
from autowsgr.utils.api_image import (
    MyTemplate,
    absolute_to_relative,
    locate_image_center,
    relative_to_absolute,
)
from autowsgr.utils.logger import Logger
from autowsgr.utils.math_functions import cal_dis


class AndroidController:
    """安卓控制器

    用于提供底层的控制接口
    """

    def __init__(
        self,
        dev: Android,
        config: UserConfig,
        logger: Logger,
    ) -> None:
        self._pool = ProcessPoolExecutor()

        self.dev: Android = dev
        self.show_android_input: bool = config.show_android_input
        self.delay: float = config.delay
        self.logger: Logger = logger
        self.screen: NDArray = self.get_raw_screen()
        self.resolution = self.screen.shape[:2][::-1]
        # (height, width, dimension) -> (width, height)
        self.logger.info(f'resolution:{self.resolution}')

    # ========= 基础命令 =========
    def shell(self, cmd):
        """向链接的模拟器发送 shell 命令
        Args:
            cmd (str):命令字符串
        """
        return self.dev.shell(cmd)

    def get_frontend_app(self):
        """获取前台应用的包名"""
        return self.shell('dumpsys window | grep mCurrentFocus')

    def start_background_app(self, package_name):
        self.dev.start_app(package_name)
        self.shell('input keyevent 3')

    def start_app(self, package_name):
        self.dev.start_app(package_name)

    def stop_app(self, package_name):
        self.dev.stop_app(package_name)

    def list_apps(self) -> bytes | str | Any:
        """列出所有正在运行的应用"""
        return self.dev.shell('ps')

    def is_game_running(self, app='zhanjian2'):
        """检查一个应用是否在运行

        Args:
            app (str, optional): 应用名, 默认为 "战舰少女R".

        Returns:
            bool:
        """
        try:
            return app in self.list_apps()  # type: ignore
        except Exception as e:
            self.logger.error(f'检查游戏是否在运行失败: {e}')
            return False

    # ========= 输入控制信号 =========
    def text(self, t):
        """输入文本
        需要焦点在输入框时才能输入
        """
        if hasattr(self, 'first_type') and self.first_type:
            self.logger.debug('第一次尝试输入, 测试中...')
            self.dev.text('T')
            time.sleep(0.5)
            self.dev.shell('input keyevent 67')
            self.first_type = False
        self.logger.debug(f'正在输入: {t}')
        self.dev.text(t)

    def relative_click(
        self,
        x: float,
        y: float,
        times: int = 1,
        delay: float = 0.5,
        enable_subprocess: bool = False,
    ) -> th.Thread | None:
        """点击模拟器相对坐标 (x,y).
        Args:
            x,y:相对坐标
            delay:点击后延时(单位为秒)
            enable_subprocess:是否启用多线程加速
            Note:
                if 'enable_subprocess' is True,arg 'times' must be 1
        Returns:
            enable_subprocess == False:None
            enable_subprocess == True:A class threading.Thread refers to this click subprocess
        """
        if self.show_android_input:
            self.logger.debug(f'click ({x:.3f} {y:.3f})')
        x, y = relative_to_absolute((x, y), self.resolution)

        if times < 1:
            raise ValueError("invalid arg 'times' " + str(times))
        if delay < 0:
            raise ValueError("arg 'delay' should be positive or 0")
        if enable_subprocess and times != 1:
            raise ValueError(
                "subprocess enabled but arg 'times' is not 1 but " + str(times),
            )
        if enable_subprocess:
            p = th.Thread(target=lambda: self.shell(f'input tap {x!s} {y!s}'))
            p.start()
            return p

        for _ in range(times):
            self.shell(f'input tap {x!s} {y!s}')
            time.sleep(delay * self.delay)
        return None

    def click(self, x, y, times=1, delay=0.3, enable_subprocess=False):
        """点击模拟器相对坐标 (x,y).
        Args:
            x,y:相对横坐标  (相对 960x540 屏幕)
            delay:点击后延时(单位为秒)
            enable_subprocess:是否启用多线程加速
            Note:
                if 'enable_subprocess' is True,arg 'times' must be 1
        Returns:
            enable_subprocess == False:None
            enable_subprocess == True:A class threading.Thread refers to this click subprocess
        """
        x, y = absolute_to_relative((x, y), (960, 540))
        self.relative_click(x, y, times, delay, enable_subprocess)

    def relative_swipe(self, x1, y1, x2, y2, duration=0.5, delay=0.5):
        """匀速滑动模拟器相对坐标 (x1, y1) 到 (x2, y2).
        Args:
            x1, y1, x2, y2: 相对坐标
            duration: 滑动总时间
            delay: 滑动后延时(单位为秒)
        """
        if delay < 0:
            raise ValueError("arg 'delay' should be positive or 0")
        x1, y1 = relative_to_absolute((x1, y1), self.resolution)
        x2, y2 = relative_to_absolute((x2, y2), self.resolution)
        duration = int(duration * 1000)
        input_str = f'input swipe {x1!s} {y1!s} {x2!s} {y2!s} {duration}'
        if self.show_android_input:
            self.logger.debug(input_str)
        self.shell(input_str)
        time.sleep(delay)

    def swipe(self, x1, y1, x2, y2, duration=0.5, delay=0.5):
        """匀速滑动模拟器相对坐标 (x1,y1) 到 (x2,y2).
        Args:
            x1, y1, x2, y2:相对坐标 (960x540 屏幕)
            duration: 滑动总时间
            delay: 滑动后延时(单位为秒)
        """
        x1, y1 = absolute_to_relative((x1, y1), (960, 540))
        x2, y2 = absolute_to_relative((x2, y2), (960, 540))
        self.relative_swipe(x1, y1, x2, y2, duration, delay)

    def relative_long_tap(self, x, y, duration=1, delay=0.5):
        """长按相对坐标 (x, y)
        Args:
            x, y: 相对坐标
            duration (int, optional): 长按时间(秒). Defaults to 1.
            delay (float, optional): 操作后等待时间(秒). Defaults to 0.5.
        """
        if delay < 0:
            raise ValueError("arg 'delay' should be positive or 0")
        if duration <= 0.2:
            raise ValueError(
                "duration time too short,arg 'duration' should greater than 0.2",
            )
        x, y = relative_to_absolute((x, y), self.resolution)
        self.swipe(x, y, x, y, duration=duration, delay=delay)

    def long_tap(self, x, y, duration=1, delay=0.5):
        """长按相对坐标 (x,y)
        Args:
            x,y: 相对 (960x540 屏幕) 横坐标
            duration (int, optional): 长按时间(秒). Defaults to 1.
            delay (float, optional): 操作后等待时间(秒). Defaults to 0.5.
        """
        x, y = absolute_to_relative((x, y), (960, 540))
        self.relative_long_tap(x, y, duration, delay)

    # ======== 屏幕相关 ========
    def get_raw_screen(self) -> NDArray[uint8]:
        """返回一个未裁剪的屏幕"""
        start_time = time.time()
        while (screen := self.dev.snapshot(quality=99)) is None:
            if time.time() - start_time > 10:
                raise CriticalErr('截图持续返回 None, 模拟器可能已经失去响应')
            time.sleep(0.1)
        return screen

    def update_screen(self):
        self.screen = self.get_raw_screen()

    def get_screen(self, resolution=(1280, 720), need_screen_shot=True):
        """获取屏幕图片

        return (ndarray): 宽x长 array[720][1280]
        """
        if need_screen_shot:
            self.update_screen()
        return cv2.resize(self.screen, resolution)

    def get_pixel(self, x, y, screen_shot=False) -> list[int]:
        """获取当前屏幕相对坐标 (x,y) 处的像素值
        Args:
            x (int): [0, 960)
            y (int): [0, 540)
        Returns:
            list[]: RGB 格式的像素值
        """
        if screen_shot:
            self.update_screen()
        if len(self.screen) != 540:
            self.screen = cv2.resize(self.screen, (960, 540))
        return [self.screen[y][x][2], self.screen[y][x][1], self.screen[y][x][0]]

    def check_pixel(
        self,
        position: tuple[int, int],
        bgr_color,
        distance=30,
        screen_shot=False,
    ) -> bool:
        r"""检查像素点是否满足要求
        Args:
            position (_type_): (x, y) 坐标, x 是长, 相对 960x540 的值, x \in [0, 960)

            bgr_color (_type_): 三元组, 顺序为 blue green red, 值为像素值

            distance (int, optional): 最大相差欧氏距离. Defaults to 30.

            screen_shot (bool, optional): 是否重新截图. Defaults to False.
        return:
            bool:是否满足要求
        """
        color = self.get_pixel(*position, screen_shot)
        color.reverse()
        return cal_dis(color, bgr_color) < distance**2

    def get_image_position(
        self,
        images: MyTemplate | list[MyTemplate],
        need_screen_shot=True,
        confidence=0.85,
    ):
        """从屏幕中找出和多张模板图像匹配度超过阈值的矩阵区域的中心坐标,如果有多个,返回第一个
        Args:
            need_screen_shot (int, optional): 是否重新截取屏幕. Defaults to 1.
        Returns:
            如果找到:返回一个二元组表示相对坐标 (相对 960x540 屏幕)

            否则返回 None
        """
        if not isinstance(images, Iterable):
            images = [images]
        if need_screen_shot:
            self.update_screen()
        for image in images:
            res = locate_image_center(self.screen, image, confidence)
            if res is not None:
                rel_pos = absolute_to_relative(res, self.resolution)
                return relative_to_absolute(rel_pos, (960, 540))
        # results = self._pool.map(partial(locate_image_center, self.screen, confidence=confidence), images)
        # for res in results:
        #     if res is not None:
        #         rel_pos = absolute_to_relative(res, self.resolution)
        #         return relative_to_absolute(rel_pos, (960, 540))
        return None

    def image_exist(
        self,
        images,
        need_screen_shot=True,
        confidence=0.85,
    ):
        """判断图像是否存在于屏幕中
        Returns:
            bool:如果存在为 True 否则为 False
        """
        if not isinstance(images, list):
            images = [images]
        if need_screen_shot:
            self.update_screen()
        return self.get_image_position(images, False, confidence) is not None

    def wait_image(
        self,
        image: MyTemplate | list[MyTemplate],
        confidence=0.85,
        timeout: float = 10,
        gap=0.15,
        after_get_delay: float = 0,
    ):
        """等待一张图片出现在屏幕中,置信度超过一定阈值(支持多图片)

        Args:
            timeout (int, optional): 最大等待时间. Defaults to 10.
        Returns:
            如果在 timeout 秒内发现,返回一个二元组表示其相对(960x540 屏幕)位置

            否则返回 False
        """
        if timeout < 0:
            raise ValueError("arg 'timeout' should at least be 0 but is ", str(timeout))
        start_time = time.time()
        while True:
            x = self.get_image_position(image, True, confidence)
            if x is not None:
                time.sleep(after_get_delay)
                return x
            if time.time() - start_time > timeout:
                time.sleep(gap)
                return False
            time.sleep(gap)

    def wait_images(
        self,
        images=None,
        confidence=0.85,
        gap=0.15,
        after_get_delay: float = 0,
        timeout: float = 10,
    ):
        """等待一系列图片中的一个在屏幕中出现

        Args:
            images (list, optional): 很多图片,可以是列表或字典. Defaults to [].
            confidence (_type_, optional): 置信度. Defaults to 0.85.
            timeout (int, optional): 最长等待时间. Defaults to 10.

        Raises:
            TypeError: image_list 中有不合法参数

        Returns:
            None: 未找到任何图片

            int: 第一个出现的图片的下标(0-based) if images is a list

            the key of the value: if images is a dict
        """
        if timeout < 0:
            raise ValueError("arg 'timeout' should at least be 0 but is ", str(timeout))
        if images is None:
            return None
        if isinstance(images, MyTemplate):
            images = [(0, images)]
        elif isinstance(images, list | tuple) and isinstance(images[0], MyTemplate):
            images = list(enumerate(images))  # 把列表转化为元组
        # TODO: 后续优化此内容，当图片结尾有数字的时候会生成一个列表，也就是列表里面嵌套列表，先前不支持此类型
        elif isinstance(images, list | tuple) and isinstance(images[0], list | tuple):
            images = list(enumerate(images))
        elif isinstance(images, dict):
            images = images.items()
        else:
            images = images.__dict__.items()

        start_time = time.time()
        while True:
            self.update_screen()
            for res, image in images:
                if self.image_exist(image, False, confidence):
                    time.sleep(after_get_delay)
                    return res
            # exists = self._pool.map(partial(self.image_exist, confidence=confidence), [image for _, image in images])
            # for (res, _), exist in zip(images, exists):
            #     if exist:
            #         time.sleep(after_get_delay)
            #         return res
            time.sleep(gap)
            if time.time() - start_time > timeout:
                return None

    def wait_images_position(
        self,
        images: list | None = None,
        confidence=0.85,
        gap=0.15,
        after_get_delay: float = 0,
        timeout: float = 10,
    ):
        """等待一些图片,并返回第一个匹配结果的位置

        参考 wait_images
        """
        if images is None:
            images = []
        if not isinstance(images, Iterable):
            images = [images]
        rank = self.wait_images(images, confidence, gap, after_get_delay, timeout)
        if rank is None:
            return None
        assert isinstance(rank, int)
        return self.get_image_position(images[rank], False, confidence)

    def click_image(self, image, must_click=False, timeout: float = 0, delay=0.5):
        """点击一张图片的中心位置

        Args:
            image (MyTemplate): 目标图片

            must_click (bool, optional): 如果为 True,点击失败则抛出异常. Defaults to False.

            timeout (int, optional): 等待延时. Defaults to 0.

            delay (float, optional): 点击后延时. Defaults to 0.5.

        Raises:
            NotFoundErr: 如果在 timeout 时间内未找到则抛出该异常
        Returns:
            bool:如果找到图片返回匹配位置，未找到则返回None
        """
        pos = self.wait_images_position(image, gap=0.03, timeout=timeout)
        if pos is None:
            if not must_click:
                return False
            raise ImageNotFoundErr(f'Target image not found:{image.filepath!s}')

        self.click(*pos, delay=delay)
        return pos

    def log_screen(
        self,
        need_screen_shot=False,
        resolution=(960, 540),
        ignore_existed_image=True,
        name=None,
    ):
        """向默认数据记录路径记录当前屏幕数据,带时间戳保存,960x540大小
        Args:
            need_screen_shot (bool, optional): 是否新截取一张图片. Defaults to False.
        """
        if need_screen_shot:
            self.update_screen()
        screen = copy.deepcopy(self.screen)
        screen = cv2.resize(screen, resolution)
        if name is None:
            self.logger.log_image(
                image=screen,
                name=datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'),
                ignore_existed_image=ignore_existed_image,
            )
        else:
            self.logger.log_image(
                image=screen,
                name=name,
                ignore_existed_image=ignore_existed_image,
            )



import threading as th
import time

from autowsgr.configs import UserConfig
from autowsgr.constants.custom_exceptions import CriticalErr, ImageNotFoundErr, NetworkErr
from autowsgr.constants.image_templates import IMG
from autowsgr.constants.other_constants import ALL_PAGES
from autowsgr.constants.ui import WSGR_UI, Node
from autowsgr.port.common import Port
from autowsgr.timer.backends import EasyocrBackend, PaddleOCRBackend
from autowsgr.timer.controllers import (
    AndroidController,
    MacController,
    OSController,
    WindowsController,
)
from autowsgr.types import OcrBackend, OSType
from autowsgr.utils.io import create_nested_dict, recursive_dict_update, yaml_to_dict
from autowsgr.utils.logger import Logger
from autowsgr.utils.operator import unzip_element


class Timer(AndroidController):
    """程序运行记录器, 用于记录和传递部分数据, 同时用于区分多开, WSGR 专用"""

    def __init__(self, config: UserConfig, logger: Logger) -> None:
        self.config = config
        self.logger = logger

        self.initialize_resources()
        self.initialize_controllers()
        self.initialize_ocr()

        # 初始化状态变量
        self.everyday_check = True
        self.ship_stats = [0, 0, 0, 0, 0, 0, 0]  # 我方舰船状态
        self.enemy_type_count = {}  # 字典,每种敌人舰船分别有多少
        self.now_page = None  # 当前所在 UI 名
        self.resources = None  # 当前四项资源量
        self.got_ship_num = 0  # 当天已掉落的船
        self.got_loot_num = 0  # 当天已掉落的胖次
        self.quick_repaired_cost = 0  # 消耗快修数量
        self.can_get_loot = False  # 是否可以捞胖次
        """
        以上时能用到的
        以下是暂时用不到的
        """
        self.last_mission_completed = 0
        self.last_expedition_check_time = time.time()
        self.port = Port(logger)
        self.init()

    def initialize_resources(self) -> None:
        # 加载资源
        if self.config.plan_root is None:
            raise ValueError('plan_root is not defined')
        self.ui = WSGR_UI
        self.plan_tree = recursive_dict_update(
            create_nested_dict(self.config.default_plan_root),
            create_nested_dict(self.config.plan_root),
        )
        self.ship_names = unzip_element(
            recursive_dict_update(
                yaml_to_dict(self.config.default_ship_name_file),
                yaml_to_dict(self.config.ship_name_file),
            ).values(),
        )
        self.logger.info('资源加载成功')

    def initialize_controllers(self) -> None:
        # 初始化操作系统控制器
        adapter_fun = {
            OSType.windows: WindowsController,
            OSType.macos: MacController,
        }
        self.os_controller: OSController = adapter_fun[self.config.os_type](
            self.config,
            self.logger,
        )
        # 初始化android控制器
        dev = self.os_controller.connect_android()
        AndroidController.__init__(self, dev, self.config, self.logger)
        self.logger.info('控制器初始化成功')

    def initialize_ocr(self) -> None:
        # 初始化 OCR 后端
        match self.config.ocr_backend:
            case OcrBackend.easyocr:
                self.ocr_backend = EasyocrBackend(self.config, self.logger)
            case OcrBackend.paddleocr:
                self.ocr_backend = PaddleOCRBackend(self.config, self.logger)
        self.logger.info('OCR 后端初始化成功')

    # ========================= OCR 功能穿透 =========================
    def recognize(
        self,
        img,
        allowlist: list[str] | None = None,
        candidates: list[str] | None = None,
        multiple=False,
        allow_nan=False,
        rgb_select=None,
        tolerance=30,
        **kwargs,
    ):
        """识别任意字符串"""
        return self.ocr_backend.recognize(
            img,
            allowlist,
            candidates,
            multiple,
            allow_nan,
            rgb_select,
            tolerance,
            **kwargs,
        )

    def recognize_number(
        self,
        img,
        extra_chars='',
        multiple=False,
        allow_nan=False,
        **kwargs,
    ):
        """识别数字"""
        return self.ocr_backend.recognize_number(
            img,
            extra_chars,
            multiple,
            allow_nan,
            **kwargs,
        )

    def recognize_ship(self, image, candidates, **kwargs):
        """传入一张图片,返回舰船信息,包括名字和舰船型号"""
        return self.ocr_backend.recognize_ship(image, candidates, **kwargs)

    # ========================= 初级游戏控制 =========================
    def init(self):
        """初始化游戏状态, 以便进一步的控制"""
        # ========== 启动游戏 ==========
        if self.config.account is not None and self.config.password is not None:
            self.restart(account=self.config.account, password=self.config.password)
        if not self.is_game_running():
            self.start_game()
        self.start_app(self.config.app_name)

        # ========== 检查游戏页面状态 ============
        try:
            self.set_page()
            if isinstance(self.now_page, Node):
                self.logger.info(f'启动成功, 当前位置: {self.now_page.name}')
            else:
                if self.config.check_page:
                    self.logger.warning('无法确定当前页面, 尝试重启游戏')
                    self.restart()
                    self.set_page()
                else:
                    self.logger.warning('在无法确定页面的情况下继续.')
        except Exception as ex:
            self.logger.warning(f'出现未知错误, 尝试重启游戏:{ex}')
            self.restart()
            self.set_page()

    def log_in(self, account, password):
        pass

    def log_out(self, account, password):
        """在登录界面登出账号"""

    def start_game(self, account=None, password=None, delay=1.0):
        """启动游戏"""
        self.start_app(self.config.app_name)
        res = self.wait_images(
            [IMG.start_image[2], *IMG.confirm_image[1:]],
            0.85,
            gap=1,
            timeout=70 * delay,
        )

        if res is None:
            raise TimeoutError('start_app timeout')
        if res != 0:
            self.confirm_operation()
            if not self.wait_image(IMG.start_image[2], timeout=200):
                raise TimeoutError('resource downloading timeout')

        # ========= 登录 =========

        if account is not None and password is not None:
            self.click(75, 450)
            if not self.wait_image(IMG.start_image[3]):
                raise TimeoutError("can't enter account manage page")
            self.click(460, 380)
            if not self.wait_image(IMG.start_image[4]):
                raise TimeoutError("can't logout successfully")
            self.click(540, 180)
            for _ in range(20):
                p = th.Thread(target=lambda: self.shell('input keyevent 67'))
                p.start()
            p.join()
            self.text(str(account))
            self.click(540, 260)
            for _ in range(20):
                p = th.Thread(target=lambda: self.shell('input keyevent 67'))
                p.start()
            p.join()
            time.sleep(0.5)
            self.text(str(password))
            self.click(400, 330)
            res = self.wait_images([IMG.start_image[5], IMG.start_image[2]])
            if res is None:
                raise TimeoutError('login timeout')
            if res == 0:
                raise CriticalErr('password or account is wrong')

        # =========== 开始游戏 ===========

        delay = 15
        while self.image_exist(IMG.start_image[2]):
            self.click_image(IMG.start_image[2])
            if self.wait_image(IMG.game_ui[3], timeout=delay, after_get_delay=0.25):
                break
            delay *= 2
            if delay > 30:
                raise ImageNotFoundErr("can't start game")
        # ============ 关闭新闻, 领取每日奖励 =============

        try:
            if self.everyday_check:
                self.logger.info('正在尝试关闭新闻, 领取奖励')
                if self.wait_image(IMG.start_image[6], timeout=2):  # 新闻与公告,设为今日不再显示
                    if not self.check_pixel((70, 485), (201, 129, 54)):
                        self.click(70, 485)
                    self.click(30, 30)
                if self.wait_image(IMG.start_image[7], timeout=7):  # 每日签到
                    self.click(474, 357)
                    self.confirm_operation(must_confirm=True, timeout=2)
                self.everyday_check = False
            self.go_main_page()
            self.logger.info('游戏启动成功!')
        except:
            raise CriticalErr('fail to start game')

    def restart(self, times=0, *args, **kwargs):
        try:
            self.shell(f'am force-stop {self.config.app_name}')
            self.shell('input keyevent 3')
            self.start_game(**kwargs)
        except Exception:
            if not self.os_controller.is_android_online():
                pass

            elif times == 1:
                raise CriticalErr('on restart,')

            elif not self.os_controller.check_network():
                for i in range(11):
                    time.sleep(10)
                    if self.os_controller.check_network():
                        break
                    if i == 10:
                        raise NetworkErr

            elif self.is_game_running():
                raise CriticalErr('CriticalErr on restart function')

            self.os_controller.connect_android()
            self.restart(times + 1, *args, **kwargs)

    def is_other_device_login(self, timeout=2):
        """检查是否有其他设备登录顶号"""
        return self.wait_images(IMG.error_image['user_remote_login'], timeout=timeout) is not None

    def process_other_device_login(self, timeout=2):
        """处理其他设备登录顶号
        TODO: 重新登录以后写，暂时留空,直接抛出错误
        """
        if self.is_other_device_login(timeout):
            self.log_screen(need_screen_shot=True, name='other device login.PNG')
            self.logger.error('other device login')
            raise CriticalErr('other device login')

    def is_bad_network(self, timeout=10):
        """检查是否为网络状况问题"""
        return (
            self.wait_images(
                [IMG.symbol_image[10]] + IMG.error_image['bad_network'],
                timeout=timeout,
            )
            is not None
        )

    def reset_chapter_map(self):
        self.port.chapter = None
        self.port.map = None

    def process_bad_network(self, extra_info='', timeout=10):
        """判断并处理网络状况问题
        Returns:
            bool: 如果为 True 则表示为网络状况问题,并已经成功处理,否则表示并非网络问题或者处理超时.
        Raise:
            TimeoutError:处理超时
        """
        start_time = time.time()
        while self.is_bad_network(timeout):
            self.log_screen(need_screen_shot=True, name='bad_network.PNG')
            self.logger.warning(f'bad network: {extra_info}')

            # 等待网络恢复

            if not self.os_controller.wait_network():
                raise NetworkErr("can't connect to www.moefantasy.com")

            # 处理网络问题
            while self.wait_images(
                [IMG.symbol_image[10]] + IMG.error_image['bad_network'],
                timeout=3,
            ):
                time.sleep(0.5)

                if self.image_exist(IMG.error_image['bad_network']):
                    self.click_image(IMG.error_image['network_retry'])

                if not self.wait_images(
                    [IMG.symbol_image[10]] + IMG.error_image['bad_network'],
                    timeout=5,
                ):
                    self.logger.debug('ok network problem solved')
                    self.reset_chapter_map()
                    return True

                if time.time() - start_time > 1800:
                    raise TimeoutError('process bad network timeout')
        return False

    # ========================= 维护当前所在游戏界面 =========================
    def _integrative_page_identify(self) -> int | None:
        positions = [(171, 47), (300, 47), (393, 47), (504, 47), (659, 47)]
        for i, position in enumerate(positions):
            if self.check_pixel(position, (225, 130, 16)):
                return i + 1
        return None

    def identify_page(self, name, need_screen_shot=True):
        if need_screen_shot:
            self.update_screen()

        if (name == 'main_page') and (self.identify_page('options_page', False)):
            return False
        if (name == 'map_page') and (
            self._integrative_page_identify() != 1 or self.check_pixel((35, 297), (47, 253, 226))
        ):
            return False
        if (name == 'build_page') and (self._integrative_page_identify() != 1):
            return False
        if (name == 'develop_page') and (self._integrative_page_identify() != 3):
            return False
        return self.image_exist(IMG.identify_images[name], False)

    def wait_pages(self, names, timeout=10, gap=0.1, after_wait=0.1):
        start_time = time.time()
        if isinstance(names, str):
            names = [names]
        while True:
            self.update_screen()
            for i, name in enumerate(names):
                if self.identify_page(name, False):
                    time.sleep(after_wait)
                    return i + 1

            if time.time() - start_time > timeout:
                break
            time.sleep(gap)

        if self.is_bad_network(timeout=3) and self.process_bad_network("can't wait pages"):
            return self.wait_pages(names, timeout, gap, after_wait)
        if self.is_other_device_login():
            self.process_other_device_login()

        raise TimeoutError(f'identify timeout of{names!s}')

    def get_now_page(self):
        """获取并返回当前页面名称"""
        self.update_screen()
        for page in ALL_PAGES:
            if self.identify_page(page, need_screen_shot=False):
                return page
        return 'unknown_page'

    def check_now_page(self):
        return isinstance(self.now_page, Node) and self.identify_page(name=self.now_page.name)

    def operate(self, end: Node):
        if not isinstance(self.now_page, Node):
            self.logger.error('now_page is not a Node object')
            raise TypeError('now_page is not a Node object')
        ui_list = self.ui.find_path(self.now_page, end)
        for next in ui_list[1:]:
            edge = self.now_page.find_edge(next)
            if edge is None:
                self.logger.error(
                    f'no edge found between {self.now_page.name} and {next.name}',
                )
                raise ValueError('no edge found')
            opers = edge.operate()
            self.now_page = next
            for oper in opers:
                fun, args = oper
                if fun == 'click':
                    self.click(*args)
                else:
                    raise ValueError(f'unknown function name: {fun}')

            if edge.other_dst is not None:
                dst = self.wait_pages(names=[self.now_page.name, edge.other_dst.name])
                if dst == 1:
                    continue
                self.logger.debug(
                    f'Go page: {self.now_page}, but arrive: {edge.other_dst.name}',
                )
                self.now_page = self.ui.get_node_by_name(
                    [self.now_page.name, edge.other_dst.name][dst - 1],
                )
                self.logger.debug(f'Now page: {self.now_page}')
                if isinstance(self.now_page, Node) and self.now_page.name == 'expedition_page':
                    try_to_get_expedition(self)
                self.operate(end)
                return
            self.wait_pages(names=[self.now_page.name])
            time.sleep(0.25)

    def set_page(self, page_name=None, page=None):
        if page_name is None and page is None:
            now_page = self.get_now_page()

            if now_page is None:
                raise ImageNotFoundErr('无法识别该页面')
            if now_page != 'unknown_page':
                self.now_page = self.ui.get_node_by_name(now_page)
            else:
                self.now_page = now_page
        elif page is not None:
            if not isinstance(page, Node):
                self.logger.error('arg:page must be an controller.ui.Node object')
                raise ValueError

            self.now_page = page if (self.ui.page_exist(page)) else 'unknown_page'
        else:
            page = self.ui.get_node_by_name(page_name)
            if page is None:
                page = 'unknown_page'

            self.now_page = page

    def walk_to(self, end, try_times=0):
        try:
            if isinstance(self.now_page, str) and 'unknown' in self.now_page:
                self.go_main_page()
            if isinstance(end, Node):
                self.operate(end)
                self.wait_pages(end.name)
                return
            if isinstance(end, str):
                end = self.ui.get_node_by_name(end)
                if end is None:
                    self.logger.error('unacceptable value of end: {end}')
                    raise ValueError('illegal value:end, in Timer.walk_to')
                self.walk_to(end)

        except TimeoutError as exception:
            if try_times > 3:
                raise TimeoutError("can't access the page")
            if self.is_other_device_login():
                self.process_other_device_login()
            if not self.is_bad_network(timeout=2):
                self.logger.debug(
                    'wrong path is operated,anyway we find a way to solve,processing',
                )
                self.logger.debug('wrong info is:', exception)
                self.go_main_page()
                self.walk_to(end, try_times + 1)
            else:
                while True:
                    if self.process_bad_network(
                        "can't walk to the position because a TimeoutError",
                    ):
                        try:
                            if not self.wait_pages(names=self.now_page.name, timeout=1):
                                self.set_page(self.get_now_page())
                        except Exception:
                            self.go_main_page()
                        else:
                            break
                    else:
                        raise ValueError('unknown error')
                self.walk_to(end)

    def go_main_page(
        self,
        quit_operation_time: int = 0,
        list1: list | None = None,
        list2: list | None = None,
    ) -> None:
        """回退到游戏主页
        Args:
            timer (Timer): _description_
            quit_operation_time (int, optional): _description_. Defaults to 0.
            list1 (list, optional): _description_. Defaults to [].
            list2 (list, optional): _description_. Defaults to [].

        Raises:
            ValueError: _description_
        """
        list1 = [] if list1 is None else list1
        list2 = [] if list2 is None else list2

        if quit_operation_time > 200:
            if self.is_other_device_login():
                self.process_other_device_login()

            if self.is_bad_network(timeout=3):
                if self.process_bad_network("can't go main page"):
                    self.go_main_page(0, list1)
                    return
            else:
                self.logger.error("Unknown error,can't go main page")
                raise ValueError("Error,Couldn't go main page")
        self.reset_chapter_map()
        self.now_page = self.ui.get_node_by_name('main_page')
        if len(list1) == 0:
            list1 = IMG.back_buttons[1:] + list2
        if list1 is None:
            raise ValueError('list1 is None')
        type = self.wait_images([*list1, IMG.game_ui[3]], 0.8, timeout=0)

        if type is None:
            self.go_main_page(quit_operation_time + 1, list1)
            return

        if type >= len(list1):
            type = self.wait_images(list1, timeout=0)
            if type is None:
                return

        pos = self.get_image_position(list1[type], False, 0.8)
        if pos is None:
            raise ImageNotFoundErr('no image found, pos is None')
        self.click(pos[0], pos[1])

        new_list = [*list1[1:], list1[0]]
        self.go_main_page(quit_operation_time + 1, new_list)

    def goto_game_page(self, target='main', extra_check=False):
        """到某一个游戏界面

        Args:
            target (str, str): 目标章节名(见 ./constants/other_constants). Defaults to 'main'.
        """
        self.walk_to(target)
        if extra_check:
            if not isinstance(self.now_page, Node):
                self.logger.error(f'now_page: {self.now_page} is not Node')
                raise TypeError('now_page is not a Node object')
            self.wait_pages(names=[self.now_page.name])

    def confirm_operation(
        self,
        must_confirm=False,
        delay=0.5,
        confidence=0.9,
        timeout=0,
    ):
        """等待并点击弹出在屏幕中央的各种确认按钮

        Args:
            must_confirm (int, optional): 是否必须按. Defaults to 0.
            delay (float, optional): 点击后延时(秒). Defaults to 0.5.
            timeout (int, optional): 等待延时(秒),负数或 0 不等待. Defaults to 0.

        Raises:
            ImageNotFoundErr: 如果 must_confirm = True 但是 timeout 之内没找到确认按钮排除该异常
        Returns:
            bool:True 为成功,False 为失败
        """
        pos = self.wait_images(
            IMG.confirm_image[1:],
            confidence=confidence,
            timeout=timeout,
        )
        if pos is None:
            if must_confirm:
                raise ImageNotFoundErr('no confirm image found')
            return False
        assert isinstance(pos, int)
        res = self.get_image_position(
            IMG.confirm_image[pos + 1],
            confidence=confidence,
            need_screen_shot=False,
        )
        if res is None:
            return False
        self.click(res[0], res[1], delay=delay)
        return True


def process_error(timer: Timer):
    """这个方法目前没有用到, ControllerAdapter也没有定义过, 暂时保留"""
    print('processing errors')
    if not timer.ControllerAdapter.is_android_online() or not timer.is_game_running():  # type: ignore
        timer.ControllerAdapter.restart_android()  # type: ignore
        timer.ControllerAdapter.connect_android()  # type: ignore

        return 'Android Restarted'

    return 'ok,bad network' if timer.process_bad_network() else 'ok,unknown error'


def try_to_get_expedition(timer: Timer):
    timer.logger.info('开始收取远征奖励...')
    get, pos = False, timer.wait_image(IMG.game_ui[6], timeout=2)
    while pos:
        timer.click(pos[0], pos[1], delay=1)
        timer.wait_image(IMG.fight_image[3], after_get_delay=0.25)
        timer.click(900, 500, delay=1)
        timer.confirm_operation(must_confirm=True, delay=0.5, confidence=0.9)
        pos, get = timer.wait_image(IMG.game_ui[6], timeout=2), True
    return get



from autowsgr.constants.other_constants import CHAPTER_NODE_COUNT
from autowsgr.fight import NormalFightPlan
from autowsgr.scripts.main import start_script


timer = start_script()
plan = NormalFightPlan(timer, 'normal_fight/7-46SS-all.yaml')
for i, nodes in enumerate(CHAPTER_NODE_COUNT):
    if i <= 0:
        continue
    for j in range(1, nodes + 1):
        plan._change_fight_map(i, j)
print('测试成功!')



"""有选择入口和常规推图的活动可以抄这个"""

import os

from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.fight.event.event import Event
from autowsgr.fight.normal_fight import NormalFightInfo, NormalFightPlan
from autowsgr.timer import Timer


NODE_POSITION = (
    None,
    (218, 123),
    (131, 343),
    (887, 188),
    (707, 436),
    (478, 216),  # 阿尔法入口用的地图
    (478, 216),  # 贝塔入口用的地图
    (478, 216),  # 西格玛入口用的地图
    (478, 216),  # 德尔塔入口用的地图
)


class EventFightPlan20230927(Event, NormalFightPlan):
    def __init__(
        self,
        timer: Timer,
        plan_path,
        from_which_entrance=None,
        fleet_id=None,
        event='20230927',
    ) -> None:
        """
        Args:
            fleet_id : 新的舰队参数, 优先级高于 plan 文件, 如果为 None 则使用计划参数.

            from_which_entrance : 指定入口, 默认为 1 表示从 alpha 进入, 如果为 2 则从 beta 进入, 以此类推， 优先级高于 plan 文件, 如果为 0 则使用计划文件的参数, 如果都没有指定, 默认从 alpha 进入
        """
        self.event_name = event
        NormalFightPlan.__init__(self, timer, plan_path, fleet_id=fleet_id)
        Event.__init__(self, timer, event)

        if from_which_entrance is not None:
            self.from_alpha = from_which_entrance

        if 'from_which_entrance' not in self.__dict__:
            self.from_which_entrance = 0

    def _load_fight_info(self):
        self.info = EventFightInfo20230927(self.timer, self.chapter, self.map)
        self.info.load_point_positions(os.path.join(MAP_ROOT, 'event', self.event_name))

    def _change_fight_map(self, chapter_id, map_id):
        """选择并进入战斗地图(chapter-map)"""
        # self.change_difficulty(chapter_id)

    def _check_entrance(self, from_which_entrance):
        # from_which_entrance为1,2,3,4
        check_position_pixel = [
            None,
            ((119, 73), (11, 30, 73)),
            ((277, 75), (12, 32, 81)),
            ((419, 74), (14, 31, 75)),
            ((56, 72), (12, 31, 73)),
        ]

        # return self.timer.check_pixel((795, 321), (254, 148, 36), screen_shot=True)
        # 第一个是xy坐标，第二个是三原色的像素
        # from_which_entrance=int(from_which_entrance)
        if from_which_entrance == 0:
            return True
        if 1 <= from_which_entrance <= 4:
            return self.timer.check_pixel(
                check_position_pixel[from_which_entrance][0],
                check_position_pixel[from_which_entrance][1],
                screen_shot=True,
            )

        raise IndexError

    def _go_fight_prepare_page(self) -> None:
        if not self.timer.image_exist(self.info.event_image[1], need_screen_shot=0):
            self.timer.click(*NODE_POSITION[self.map])

        # 选择入口
        if not self._check_entrance(self.from_which_entrance):
            entrance_position = [None, (131, 58), (280, 54), (421, 55), (593, 51)]
            self.timer.click(*entrance_position[self.from_which_entrance])

        self.timer.wait_image(self.event_image[1])
        self.timer.click(850, 490)
        try:
            self.timer.wait_pages('fight_prepare_page', after_wait=0.15)
        except:
            self.timer.logger.warning('匹配战斗准备页面失败，尝试重新匹配')
            self.timer.go_main_page()
            self._go_map_page()
            self._go_fight_prepare_page()


class EventFightInfo20230927(Event, NormalFightInfo):
    def __init__(self, timer: Timer, chapter_id, map_id, event='20230927') -> None:
        NormalFightInfo.__init__(self, timer, chapter_id, map_id)
        Event.__init__(self, timer, event)
        self.map_image = self.common_image['special'] + [self.event_image[1]]
        self.end_page = 'unknown_page'
        self.state2image['map_page'] = [self.map_image, 5]



import os
import sys


sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from tools.ship_name import get_allow, recognize

from autowsgr.constants.other_constants import ALL_SHIP_TYPES_CN, CN_TYPE_TO_EN_TYPE


DEBUG = False


def split_str(str, keywords):
    res = []
    while len(str):
        flag = 0
        for word in keywords:
            if word == str[: len(word)]:
                res.append(word)
                str = str[len(word) :]
                flag = 1
        if flag == 0:
            str = str[1:]
    return res


def recognize_decisive_enemy(image):
    allow = get_allow(ALL_SHIP_TYPES_CN)
    char = 'A'
    for box in recognize(image, char_list=allow + '/'):
        try:
            text = box[1]
            if DEBUG:
                print(text)
            assert isinstance(text, str)
            enemy = split_str(text, keywords=ALL_SHIP_TYPES_CN)
            for i in range(len(enemy)):
                enemy[i] = CN_TYPE_TO_EN_TYPE[enemy[i]]
            print(f'{char} : {enemy}')
        except Exception as e:
            print(e)
            print(char)
        char = chr(ord(char) + 1)


if __name__ == '__main__':
    print('Please Input The Image File:')
    path = input()
    recognize_decisive_enemy(path)



import time

from autowsgr.constants.custom_exceptions import ImageNotFoundErr, ShipNotFoundErr
from autowsgr.constants.image_templates import IMG, MyTemplate
from autowsgr.constants.positions import BLOOD_BAR_POSITION
from autowsgr.game.get_game_info import check_support_stats, detect_ship_stats
from autowsgr.timer import Timer
from autowsgr.types import DestroyShipWorkMode, RepairMode, ShipType
from autowsgr.utils.api_image import absolute_to_relative, crop_image


def get_ship(timer: Timer):
    """获取掉落舰船"""

    def recognize_get_ship(timer: Timer):
        """识别获取 舰船/装备 页面斜着的文字，对原始图片进行旋转裁切"""
        NAME_POSITION = [(0.754, 0.268), (0.983, 0.009), 25]
        ship_name_recognize_result = timer.recognize(
            crop_image(timer.screen, *NAME_POSITION),
            candidates=timer.ship_names,
        )
        assert ship_name_recognize_result is not None
        ship_name = ship_name_recognize_result[1]

        TYPE_POSITION = [(0.79, 0.29), (0.95, 0.1), 25]
        ship_type_recognize_result = timer.recognize(crop_image(timer.screen, *TYPE_POSITION))
        # 因为 allow_nan 为 False, 所以肯定不是 None
        assert ship_type_recognize_result is not None
        ship_type = ship_type_recognize_result[1]

        return ship_name, ship_type

    ship_name = None
    ship_type = None
    timer.got_ship_num += 1
    if timer.port.ship_factory.capacity is not None:
        timer.logger.info(
            f'当前船坞容量 {timer.port.ship_factory.occupation}/{timer.port.ship_factory.capacity}',
        )
        timer.port.ship_factory.occupation += 1
    symbol_images: list[MyTemplate] = [IMG.symbol_image[8], IMG.symbol_image[13]]
    while timer.wait_image(symbol_images, timeout=1):
        try:
            ship_name, ship_type = recognize_get_ship(timer)
            timer.logger.info(f'获取舰船: {ship_name} {ship_type}')
        except Exception as e:
            print(e)
        timer.click(915, 515, delay=0.5, times=1)
        timer.confirm_operation()
    return ship_name, ship_type


def match_night(timer: Timer, is_night):
    """匹配夜战按钮并点击"""
    timer.wait_image(IMG.fight_image[6])
    while timer.wait_image(IMG.fight_image[6], timeout=0.5):
        if is_night:
            timer.click(325, 350, delay=0.5, times=1)
        else:
            timer.click(615, 350, delay=0.5, times=1)


def click_result(timer: Timer, max_times=1):
    """点击加速两页战果界面"""
    timer.wait_images(IMG.fight_image[14])
    looted = False
    while timer.wait_image(IMG.fight_image[14], timeout=0.5):
        if timer.can_get_loot and not looted:
            if timer.image_exist(IMG.fight_result['LOOT'], need_screen_shot=True):
                timer.got_loot_num += 1
                timer.logger.info(f'捞到胖次! 当前胖次数:{timer.got_loot_num}')
                looted = True
            else:
                timer.logger.debug('没有识别到胖次')
        timer.click(915, 515, delay=0.25, times=1)


def destroy_ship(timer: Timer, ship_types: list[ShipType] | None = None):
    """解装舰船，目前仅支持：全部解装+保留装备
    Args:
        timer (Timer): _description_
        ship_types (list[ShipType], optional): Override Config 里面的解装舰船类型. Defaults to None. 若为 None 则使用 Config 中的配置
    """

    timer.go_main_page()
    timer.goto_game_page('destroy_page')
    timer.set_page('destroy_page')

    timer.click(90, 206, delay=1.5)  # 点添加

    # 选择舰船类型
    if timer.config.destroy_ship_work_mode is not DestroyShipWorkMode.disable:
        destroy_types = ship_types if ship_types is not None else timer.config.destroy_ship_types

        if timer.config.destroy_ship_work_mode is DestroyShipWorkMode.exclude:
            intended_destroy_types: list[ShipType] = [
                x for x in ShipType.enum() if x not in destroy_types
            ]
        elif timer.config.destroy_ship_work_mode is DestroyShipWorkMode.include:
            intended_destroy_types = destroy_types
        else:
            raise ValueError('不支持的解装模式')

        if intended_destroy_types is not None:
            timer.relative_click(0.912, 0.681)
            for ship_type in intended_destroy_types:
                timer.relative_click(*ship_type.relative_position_in_destroy, times=1, delay=0.8)
            timer.relative_click(0.9, 0.85, delay=1.5)

    timer.relative_click(0.91, 0.3, delay=1.5)  # 快速选择
    timer.relative_click(0.915, 0.906, delay=1.5)  # 确定
    timer.relative_click(0.837, 0.646, delay=1.5)  # 卸下装备
    timer.relative_click(0.9, 0.9, delay=1.5)  # 解装
    timer.relative_click(0.38, 0.567, delay=1.5)  # 四星确认

    # 识别船坞容量
    timer.click(90, 206, delay=1.5)  # 点添加
    # occupation, capacity = (crop_rectangle_relative(timer.get_screen(), 0.873, 0.035, 0.102, 0.038))
    # timer.port.ship_factory.update_capacity(capacity, occupation)
    # timer.logger.info(f"舰船容量: {occupation}/{capacity}")
    timer.go_main_page()


def verify_team(timer: Timer):
    """检验目前是哪一个队伍(1~4)
    含网络状况处理
    Args:
        timer (Timer): _description_

    Raises:
        ImageNotFoundErr: 未找到队伍标志
        ImageNotFoundErr: 不在相关界面

    Returns:
        int: 队伍编号(1~4)
    """
    if not timer.identify_page('fight_prepare_page'):
        raise ImageNotFoundErr('not on fight_prepare_page')

    for _ in range(5):
        for i, position in enumerate([(64, 83), (186, 83), (310, 83), (430, 83)]):
            if timer.check_pixel(position, bgr_color=(228, 132, 16)):
                return i + 1
        time.sleep(0.2)
        timer.update_screen()

    if timer.process_bad_network():
        return verify_team(timer)

    timer.log_screen()
    raise ImageNotFoundErr


def move_team(timer: Timer, target, try_times=0):
    """切换队伍
    Args:
        timer (Timer): _description_
        target (_type_): 目标队伍
        try_times: 尝试次数
    Raise:
        ValueError: 切换失败
        ImageNotFoundErr: 不在相关界面
    """
    if try_times > 3:
        raise ValueError("can't change team")

    if not timer.identify_page('fight_prepare_page'):
        timer.log_screen()
        raise ImageNotFoundErr("not on 'fight_prepare_page' ")

    if verify_team(timer) == target:
        return
    timer.logger.info('正在切换队伍到:' + str(target))
    timer.click(110 * target, 81)
    if verify_team(timer) != target:
        move_team(timer, target, try_times + 1)


def set_support(timer: Timer, target, try_times=0):
    """启用战役支援
    Args:
        timer (Timer): _description_
        target (bool, int): 目标状态
    Raise:
        ValueError: 未能成功切换战役支援状态
    """
    target = bool(target)
    timer.goto_game_page('fight_prepare_page')

    if check_support_stats(timer) != target:
        timer.click(628, 82, delay=1)
        timer.click(760, 273, delay=1)
        timer.click(480, 270, delay=1)
        timer.logger.info('开启支援状态成功')

    if timer.is_bad_network(0) or check_support_stats(timer) != target:
        if timer.process_bad_network('set_support'):
            set_support(timer, target)
        else:
            raise ValueError("can't set right support")


def quick_repair(
    timer: Timer,
    repair_mode: int | RepairMode | list[RepairMode] = RepairMode.severe_damage,
    ship_stats=None,
    switch_back=False,
    *args,
    **kwargs,
):
    """战斗界面的快速修理
    Args:
        timer (Timer): _description_
        repair_mode: int
            1: 快修, 修中破
            2: 快修, 修大破
            3: 快修, 修复所有正在修理中的船
    """
    repair_mode_list: list[RepairMode] = []
    if isinstance(repair_mode, int) and repair_mode in [1, 2, 3]:
        repair_mode = RepairMode(repair_mode)
    try:
        if ship_stats is None:
            ship_stats = detect_ship_stats(timer)
        if not any(x in ship_stats for x in [0, 1, 2]):
            time.sleep(1)
            ship_stats = detect_ship_stats(timer)
        if not any(x in ship_stats for x in [0, 1, 2]):
            timer.logger.warning('执行修理操作时没有成功检测到舰船')
            raise ValueError('没有成功检测到舰船，请检查是否正确编队')

        if isinstance(repair_mode, RepairMode):
            repair_mode_list = [repair_mode for _ in range(6)]
        else:
            repair_mode_list = repair_mode
        assert len(repair_mode_list) == 6

        need_repair = [False for _ in range(6)]
        for i, x in enumerate(repair_mode_list):
            assert x in [RepairMode.moderate_damage, RepairMode.severe_damage, RepairMode.repairing]
            if x is RepairMode.moderate_damage:
                need_repair[i] = ship_stats[i + 1] in [1, 2, 3]
            elif x is RepairMode.severe_damage:
                need_repair[i] = ship_stats[i + 1] in [2, 3]
            # 好像修理中的是直接修, 无需通过这个逻辑
            # elif x is RepairMode.repairing:
            #     need_repair[i] = ship_stats[i + 1] in [3]
            else:
                need_repair[i] = False

        if timer.config.debug:
            timer.logger.debug('ship_stats:', ship_stats)
        if any(need_repair) or timer.image_exist(IMG.repair_image[1]):
            if timer.config.repair_manually:
                timer.logger.info('需要手动修理舰船')
                raise BaseException('需要手动修理舰船')
            timer.click(420, 420, times=2, delay=0.8)
            # 快修已经开始泡澡的船
            pos = timer.get_image_position(IMG.repair_image[1])
            while pos is not None:
                timer.port.bathroom.available_time = None
                timer.click(pos[0], pos[1], delay=1)
                pos = timer.get_image_position(IMG.repair_image[1])
            # 按逻辑修理
            for i in range(1, 7):
                if need_repair[i - 1]:
                    timer.logger.info('WorkInfo:' + str(kwargs))
                    timer.logger.info(str(i) + ' Repaired')
                    timer.quick_repaired_cost += 1
                    timer.click(
                        BLOOD_BAR_POSITION[0][i][0],
                        BLOOD_BAR_POSITION[0][i][1],
                        delay=1.5,
                    )
            if switch_back:
                timer.click(200, 420, times=2, delay=1.5)  # 返回正常切换页面
    except AssertionError:
        raise ValueError(f'修理舰船的参数不合法, 请检查你的参数:{repair_mode}')


def get_rewards(timer: Timer):
    """检查任务情况,如果可以领取奖励则领取"""
    timer.go_main_page()
    if not timer.check_pixel((694, 457), bgr_color=(45, 89, 255)):
        return 'no'
    timer.goto_game_page('mission_page')
    if timer.click_image(IMG.game_ui[15]):
        timer.relative_click(0.5, 0.5)
        timer.confirm_operation(must_confirm=True, timeout=5)
        return 'ok'
    if timer.click_image(IMG.game_ui[12]):
        timer.confirm_operation(must_confirm=True)
        return 'ok'
    return 'no'
    # timer.click(774, 502)


def repair_by_bath(timer: Timer):
    """使用浴室修理修理时间最长的单位
    Args:
        timer (Timer): _description_
    """
    timer.goto_game_page('choose_repair_page')
    timer.click(115, 233)
    if not timer.identify_page('choose_repair_page'):
        if timer.identify_page('bath_page'):
            timer.set_page('bath_page')
        else:
            timer.set_page()


def set_auto_supply(timer: Timer, type=1):
    """设置自动补给
    Args:
        timer (Timer): _description_
        type (int, optional): 1: 自动补给, 0: 不自动补给. Defaults to 1.
    """
    timer.update_screen()
    NowType = int(timer.check_pixel((48, 508), (224, 135, 35)))
    if NowType != type:
        timer.click(44, 503, delay=0.33)


def supply(timer: Timer, ship_ids=None, try_times=0):
    """补给指定舰船

    Args:
        timer (Timer): _description_
        ship_ids (list, optional): 补给舰船列表, 可以为单个整数表示单艘舰船. Defaults to [1, 2, 3, 4, 5, 6].

    Raises:
        ValueError: 补给失败
        TypeError: ship_ids 参数有误
    """
    if ship_ids is None:
        ship_ids = [1, 2, 3, 4, 5, 6]
    if try_times > 3:
        raise ValueError("can't supply ship")

    if isinstance(ship_ids, int):
        ship_ids = [ship_ids]

    timer.click(293, 420)
    for x in ship_ids:
        if not isinstance(x, int):
            raise TypeError('ship must be represent as a int but get' + str(ship_ids))
        timer.click(110 * x, 241)

    if timer.is_bad_network(0):
        timer.process_bad_network('supply ships')
        supply(timer, ship_ids, try_times + 1)


def change_ship(
    timer: Timer,
    fleet_id,
    ship_id: int,
    name=None,
    pre=None,
    ship_stats=None,
):
    """切换舰船(不支持第一舰队)"""
    if fleet_id is not None and not timer.identify_page('fight_prepare_page'):
        timer.goto_game_page('fight_prepare_page')
        move_team(timer, fleet_id)
        if fleet_id >= 5:
            # 切换为预设编队
            # 暂不支持
            return

    # 切换单船
    if ship_stats is None:
        ship_stats = detect_ship_stats(timer)
    if name is None and ship_stats[ship_id] == -1:
        return
    timer.click(110 * ship_id, 250, delay=0)
    res = timer.wait_images(
        [IMG.choose_ship_image[1], IMG.choose_ship_image[2]],
        after_get_delay=0.4,
        gap=0,
    )
    if res == 1:
        timer.click(839, 113)

    if name is None:
        timer.click(83, 167, delay=0)
        timer.wait_pages('fight_prepare_page', gap=0)
        return

    timer.click(700, 30, delay=0)
    # 等待输入框出现
    start_time = time.time()
    while not timer.check_pixel((955, 500), (255, 255, 255), screen_shot=True):
        time.sleep(0.2)
        if time.time() - start_time > 10:
            timer.log_screen(True)
            raise TimeoutError('等待输入框出现时超时')
    timer.logger.debug('输入框已经出现')
    # 输入名字检索
    timer.text(name)
    # 输入后随便点击获得检索结果
    timer.click(50, 50, delay=0.5)
    time.sleep(0.5)
    # OCR识别舰船
    if name not in timer.ship_names:
        timer.ship_names.append(name)
    ship_info = timer.recognize_ship(timer.get_screen()[:, :1048], timer.ship_names)

    # 查找识别结果中要选的舰船
    found_ship = next((ship for ship in ship_info if ship[1] == name), None)
    # 点击舰船
    if found_ship is None:
        timer.logger.error(f"Can't find ship {name}, ocr result:{ship_info}")
        if len(ship_info) == 0:
            timer.logger.warning('无法查找到任何舰船, 请检查舰船名是否有错误')
            raise ShipNotFoundErr
        timer.logger.debug('Try to click the first ship')
        if ship_stats[ship_id] == -1:
            timer.click(83, 167, delay=0)
        else:
            timer.click(183, 167, delay=0)
    else:
        center = found_ship[0]
        rel_center = absolute_to_relative(center, timer.resolution)
        timer.relative_click(*rel_center)

    timer.wait_pages('fight_prepare_page', gap=0)


def change_ships(timer: Timer, fleet_id, ship_names):
    """更换编队舰船

    Args:
        fleet_id (int): 2~4,表示舰队编号

        ship_names (舰船名称列表):

    For example:
        change_ships(timer, 2, [None, "萤火虫", "伏尔塔", "吹雪", "明斯克", None, None])

    """
    timer.logger.info(f'Change Fleet {fleet_id} to {ship_names}')
    if fleet_id is not None and not timer.identify_page('fight_prepare_page'):
        timer.goto_game_page('fight_prepare_page')
        move_team(timer, fleet_id)
    if fleet_id == 1:
        raise ValueError('change member of fleet 1 is unsupported')
    ship_stats = detect_ship_stats(timer)
    for i in range(1, 7):
        if ship_stats[i] != -1:
            change_ship(timer, fleet_id, 1, None, ship_stats=ship_stats)
    ship_names = ship_names + [None] * 6
    for i in range(len(ship_names)):
        if ship_names[i] == '':
            ship_names[i] = None
    time.sleep(0.3)
    ship_stats = detect_ship_stats(timer)
    for i in range(1, 7):
        change_ship(timer, fleet_id, i, ship_names[i], ship_stats=ship_stats)


def get_new_things(timer: Timer, lock=0):
    pass


# 是否强制点击
def cook(timer: Timer, position: int, force_click=False):
    """食堂做菜
    Args:
        position (int): 第几个菜谱
        force_click (bool, optional): 当有菜正在生效时是否继续做菜. Defaults to False.
    """
    if position < 1 or position > 3:
        raise ValueError(f'不支持的菜谱编号:{position}')
    POSITION = [None, (318, 276), (420, 140), (556, 217)]
    timer.goto_game_page('canteen_page')
    timer.click(*POSITION[position])
    try:
        timer.click_image(IMG.restaurant_image.cook, timeout=7.5, must_click=True)
        if timer.image_exist(IMG.restaurant_image.have_cook):
            timer.logger.info('当前菜的效果正在生效')
            if force_click:
                timer.relative_click(0.414, 0.628)
                if timer.image_exist(IMG.restaurant_image.no_times):
                    timer.logger.info('今日用餐次数已经用尽')
                    timer.relative_click(0.788, 0.207)
                    return False
                timer.logger.info('做菜成功')
            else:
                timer.relative_click(0.65, 0.628)
                timer.logger.info('取消做菜')
                timer.relative_click(0.788, 0.207)
        return True

    except:
        timer.logger.warning(
            f'不支持的菜谱编号:{position}, 请检查该菜谱是否有效, 或者检查今日用餐次数是否已经用尽',
        )
        return False



import time
from typing import Any

from autowsgr.constants.custom_exceptions import ImageNotFoundErr
from autowsgr.constants.image_templates import IMG
from autowsgr.timer import Timer
from autowsgr.utils.math_functions import cal_dis


class Event:
    def __init__(self, timer: Timer, event_name: str) -> None:
        self.timer = timer
        self.logger = timer.logger

        self.event_image = IMG.event[event_name]
        self.common_image = IMG.event['common']
        self.enemy_image = IMG.event['enemy']

        self.common_image['monster'] = [
            self.common_image.little_monster,
            self.common_image.big_monster,
        ]

    def _go_map_page(self):
        self.timer.go_main_page()
        self.timer.click(849, 261)

    def get_difficulty(self):
        """获取难度信息
        Returns:
            简单 0,困难 1
        这里同时有检查 _go_map_page 是否成功的功能
        如果未能检测到难度图标，但是检测到进入活动地图，默认没有通过简单难度，返回简单 0.
        """
        res = self.timer.wait_images(self.common_image.hard + self.common_image.easy)
        if res is None:
            self.logger.warning('ImageNotFoundErr: difficulty image not found')
            if self.timer.wait_image(self.event_image[2]):
                self.logger.info(
                    '成功进入活动页面，未检测到切换难度图标，请检查是否通关简单难度',
                )
                return 0
            self.timer.log_screen()
            raise ImageNotFoundErr

        if self.timer.image_exist(self.common_image.hard, need_screen_shot=False):
            return 0
        return 1

    def change_difficulty(self, chapter, retry=True) -> Any | None:
        r_difficulty = int(chapter in 'Hh')
        difficulty = self.get_difficulty()

        if r_difficulty != difficulty:
            time.sleep(0.2)
            if int(chapter in 'Hh'):
                if not self.timer.click_image(self.common_image.hard):
                    self.logger.error('请检查是否通关简单难度')
                    raise ImageNotFoundErr
            else:
                self.timer.click_image(self.common_image.easy)

            if self.get_difficulty() != r_difficulty:
                if retry:
                    return self.change_difficulty(chapter, False)
                self.timer.log_screen()
                raise ImageNotFoundErr
        return None


class PatrollingEvent(Event):
    """巡戈作战活动 类"""

    def __init__(self, timer: Timer, event_name, map_positions) -> None:
        """
        Args:
            map_positions : 从主页面点进活动后, 去到对应地图需要点的位置
                : 对于 E1~E3/H1~H3, 值为地图页面滑到最左边时的点击位置
                : 对于 E4~E6/H4~H6, 值为地图页面滑到最右边时点击的位置
                : map_positions[0] 为 None
                : map_positions[1] 为 E1 的点击位置
                : map_positions[2] 为 E2 的点击位置...
        """
        self.MAP_POSITIONS = map_positions
        super().__init__(timer, event_name)

    def enter_map(self, chapter, map):
        """从活动地图选择界面进入到巡游地图"""
        assert chapter in 'HEhe'
        assert map in range(1, 7)
        self.change_difficulty(chapter)
        if map <= 3:
            self.timer.swipe(100, 300, 600, 300, duration=0.4, delay=0.15)
            self.timer.swipe(100, 300, 600, 300, duration=0.4, delay=0.15)
        else:
            self.timer.swipe(600, 300, 100, 300, duration=0.4, delay=0.15)
            self.timer.swipe(600, 300, 100, 300, duration=0.4, delay=0.15)
        self.timer.click(*self.MAP_POSITIONS[map], delay=0.25)
        assert self.timer.wait_image(self.event_image[2]) is not False  # 是否成功进入地图

    def go_fight_prepare_page(self):
        self.timer.click(789, 455)
        assert self.timer.wait_image(IMG.identify_images['fight_prepare_page']) is not False

    def random_walk(self):
        """随机游走,寻找敌人"""
        ways = ((0, 1), (0, -1), (1, 0), (-1, 0), (1, 1), (1, -1), (-1, 1), (-1, -1))
        import random

        way = random.choice(ways)
        position = (480, 270)
        step = (320, 180)
        end = (position[0] + step[0] * way[0], position[1] + step[1] * way[1])
        self.timer.click(*end, delay=3)
        if self.timer.image_exist(self.event_image[1]):
            self.timer.click(911, 37)
        if self.timer.image_exist(self.event_image[3]):
            self.timer.click(30, 50)

    def get_close(self, images):
        while True:
            ret = self.timer.wait_images_position(
                images,
                confidence=0.8,
                gap=0.03,
                timeout=1,
            )
            if cal_dis([ret[0]], [480]) ** 0.5 < 320 and cal_dis([ret[1]], [270]) ** 0.5 < 180:
                return ret
            ret = (ret[0] - 130, ret[1]) if ret[0] > 480 else (ret[0] + 130, ret[1])
            self.timer.click(*ret)

    def find(self, images, max_times=20):
        for _ in range(max_times):
            ret = self.timer.wait_images_position(
                images,
                confidence=0.75,
                gap=0.03,
                timeout=1,
            )
            if ret is not None:
                return ret
            self.random_walk()
        return None



import copy
import datetime
import os
import warnings
from dataclasses import KW_ONLY, asdict, dataclass, field, fields
from typing import Any, Literal, TextIO
from typing_extensions import Self

import rich

from autowsgr.constants.data_roots import DATA_ROOT, OCR_ROOT
from autowsgr.types import (
    DestroyShipWorkMode,
    EmulatorType,
    FightCondition,
    Formation,
    GameAPP,
    OcrBackend,
    OSType,
    RepairMode,
    ShipType,
)


@dataclass(frozen=True)
class BaseConfig:
    _: KW_ONLY

    @classmethod
    def from_dict(cls, data: dict[str, Any], /, **kwargs: Any) -> Self:
        """Create a new instance from a dictionary."""
        data, kwargs = {**data, **kwargs}, {}
        for f in fields(cls):
            if f.name in data and f.init:
                if f.name in ATTRIBUTE_RECURSIVE:
                    if f.type is list:
                        kwargs[f.name] = [
                            ATTRIBUTE_RECURSIVE[f.name].from_dict(d) for d in data.pop(f.name)
                        ]
                    else:
                        kwargs[f.name] = ATTRIBUTE_RECURSIVE[f.name].from_dict(data.pop(f.name))
                else:
                    kwargs[f.name] = data.pop(f.name)

        self = cls(**kwargs)
        sentinel = object()
        for key, value in data.items():
            if key in ATTRIBUTE_IGNORE:
                continue
            self_value = getattr(self, key, sentinel)
            if self_value is sentinel:
                warnings.warn(f'Unexpected key: {key!r}', stacklevel=2)
                object.__setattr__(self, key, value)
            elif self_value != value:
                raise ValueError(
                    f'Conflicting values for key: {key!r}, got {value!r} and already had {self_value!r}.',
                )

        return self

    def asdict(self) -> dict[str, Any]:
        """Convert the instance to a dictionary."""
        return asdict(self)

    def __replace__(self, **kwargs: Any) -> Self:
        """Replace the attributes with the given values."""
        replaced = copy.copy(self)
        for key, value in kwargs.items():
            object.__setattr__(replaced, key, value)
        return replaced

    def pprint(self, file: TextIO | None = None) -> None:
        """Pretty print the instance."""
        rich.print(self, file=file)


@dataclass(frozen=True)
class DailyAutomationConfig(BaseConfig):
    # Routine
    auto_expedition: bool = True
    """自动重复远征"""
    auto_gain_bonus: bool = True
    """当有任务完成时自动点击"""
    auto_bath_repair: bool = True
    """空闲时自动澡堂修理"""
    auto_set_support: bool = False
    """自动开启战役支援"""
    auto_battle: bool = True
    """自动打完每日战役次数"""
    battle_type: Literal[
        '简单航母',
        '简单潜艇',
        '简单驱逐',
        '简单巡洋',
        '简单战列',
        '困难航母',
        '困难潜艇',
        '困难驱逐',
        '困难巡洋',
        '困难战列',
    ] = '困难潜艇'
    """打哪个战役"""
    auto_exercise: bool = True
    """自动打完每日的三次演习"""
    exercise_fleet_id: int | None = None
    """演习出征舰队"""

    # 常规战
    auto_normal_fight: bool = True
    """按自定义任务进行常规战"""
    normal_fight_tasks: list[str] = field(default_factory=list)
    """常规战任务列表"""
    quick_repair_limit: int | None = None
    """快修消耗上限，达到时终止常规战出征。不填则无上限"""
    stop_max_ship: bool = False
    """是否获取完当天上限500船后终止常规战出征"""
    stop_max_loot: bool = False
    """是否获取完当天上限50胖次后终止常规战出征"""


@dataclass(frozen=True)
class DecisiveBattleConfig(BaseConfig):
    chapter: int = 6
    """决战章节,请保证为 [1, 6] 中的整数. Defaults to 6."""
    level1: list[str] | None = None
    """一级舰队"""
    level2: list[str] | None = None
    """二级舰队"""
    flagship_priority: list[str] | None = None
    """旗舰优先级队列"""
    repair_level: int = 1
    """维修策略, 1 为中破修, 2 为大破修"""
    full_destroy: bool = False
    """是否船舱满了解装舰船（仅限决战）"""
    useful_skill: bool = False
    """充分利用技能, 开启时要求地图1必须为Lv1+Lv2中的船; 其余地图至少一半的船为Lv1中的船"""
    useful_skill_strict: bool = False
    """严格利用技能, 开启时要求地图1技能不能获取+1的船; useful_skill为True时本设置才生效"""

    def __post_init__(self) -> None:
        assert 1 <= self.chapter <= 6, '决战章节必须为 [1, 6] 中的整数'
        if self.level1 is None:
            object.__setattr__(
                self,
                'level1',
                ['鲃鱼', 'U-1206', 'U-47', '射水鱼', 'U-96', 'U-1405'],
            )
        if self.level2 is None:
            object.__setattr__(self, 'level2', ['U-81', '大青花鱼'])
        if self.flagship_priority is None:
            object.__setattr__(self, 'flagship_priority', ['U-1405', 'U-47', 'U-96', 'U-1206'])


@dataclass(frozen=True)
class UserConfig(BaseConfig):
    # 系统
    os_type: OSType = field(init=False)
    """操作系统类型。自动设置"""

    # 模拟器
    emulator_type: EmulatorType = EmulatorType.leidian
    """模拟器类型。mumu模拟器现在截图效率较低，后续会进行优化"""
    emulator_name: str | None = None
    """模拟器链接地址。不填则自动设置"""
    emulator_start_cmd: str | None = None
    """模拟器exe地址。如果留空则自动从注册表中查询。示例: C:/leidian/LDPlayer9/dnplayer.exe"""
    emulator_process_name: str | None = None
    """模拟器进程名。不填则自动设置"""

    # 游戏
    game_app: GameAPP = GameAPP.official
    """游戏版本。"""
    app_name: str = field(init=False)
    """游戏应用名。自动设置"""
    account: str | None = None
    """游戏账号"""
    password: str | None = None
    """游戏密码"""

    # 脚本行为
    ocr_backend: OcrBackend = OcrBackend.easyocr
    """文字识别后端。TODO: 暂时仅easyocr没问题"""
    delay: float = 1.5
    """延迟时间基本单位，单位为秒。如果模拟器卡顿可调高"""
    check_update: bool = True
    """脚本自动检查更新"""
    check_page: bool = True
    """是否在启动时检查游戏页面"""
    dock_full_destroy: bool = True
    """船坞已满时自动清空, 若设置为false则船坞已满后终止所有常规出征任务"""
    repair_manually: bool = False
    """是否手动修理, 若设置为True则需要修理时不使用快修, 结束脚本"""
    bathroom_feature_count: int = 1
    """浴室数(购买的浴室装饰数, 最大为 3) TODO: 可自动获取"""
    bathroom_count: int = 2
    """修理位置总数(最大为 12) TODO: 可自动获取"""
    default_plan_root: str = field(init=False)
    """默认计划文件根目录。"""
    plan_root: str | None = None
    """计划文件根目录。可部分覆盖default_plan_root中的文件"""
    default_ship_name_file: str = field(init=False)
    """默认舰船名文件。"""
    ship_name_file: str | None = None
    """舰船名文件。不填写则使用default_ship_name_file"""
    destroy_ship_work_mode: DestroyShipWorkMode = DestroyShipWorkMode.disable
    """解装舰船的工作模式. disable 是不启用舰种分类, include 为只解装指定舰种, exclude 为解装除指定舰种外的所有舰种"""
    destroy_ship_types: list[ShipType] = field(default_factory=list)
    """指定舰种, 参照 autowsgr/types.py 中 #191 行的 ShipType, 使用中文"""

    # Log
    log_root: str = 'log'
    """日志保存根目录"""
    log_dir: str = field(init=False)
    """日志保存路径。自动创建日期文件夹。"""
    debug: bool = True
    """是否开启调试模式，如果为 True, 则会输出更多的调试信息。"""
    log_level: Literal['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'] = 'DEBUG'
    """调试模式log_level应该设置为DEBUG"""
    show_map_node: bool = False
    """是否显示地图节点信息。"""
    show_android_input: bool = True
    """是否显示 Android 输入的信息。"""
    show_enemy_rules: bool = True
    """是否显示敌人规则。"""
    show_fight_stage: bool = True
    """是否显示战斗阶段。"""
    show_chapter_info: bool = True
    """是否显示章节信息。"""
    show_match_fight_stage: bool = True
    """是否显示匹配战斗阶段。"""
    show_decisive_battle_info: bool = True
    """是否显示决战信息。"""
    show_ocr_info: bool = True
    """是否显示OCR信息。"""

    # 嵌套的下层配置
    daily_automation: DailyAutomationConfig | None = None
    """日常自动化配置"""
    decisive_battle: DecisiveBattleConfig | None = None
    """决战自动化配置"""

    def __post_init__(self) -> None:
        # 确保类型正确
        object.__setattr__(self, 'emulator_type', EmulatorType(self.emulator_type))
        object.__setattr__(self, 'game_app', GameAPP(self.game_app))
        object.__setattr__(self, 'ocr_backend', OcrBackend(self.ocr_backend))
        object.__setattr__(
            self,
            'destroy_ship_work_mode',
            DestroyShipWorkMode(self.destroy_ship_work_mode),
        )
        if self.destroy_ship_types is None:
            object.__setattr__(self, 'destroy_ship_types', [])
        else:
            object.__setattr__(
                self,
                'destroy_ship_types',
                [ShipType(t) for t in self.destroy_ship_types],
            )

        # 系统
        object.__setattr__(self, 'os_type', OSType.auto())

        # 模拟器
        if self.emulator_name is None:
            object.__setattr__(
                self,
                'emulator_name',
                self.emulator_type.default_emulator_name(self.os_type),
            )
        if self.emulator_start_cmd is None:
            object.__setattr__(
                self,
                'emulator_start_cmd',
                self.emulator_type.auto_emulator_path(self.os_type),
            )
        assert self.emulator_start_cmd is not None
        if self.emulator_process_name is None:
            object.__setattr__(
                self,
                'emulator_process_name',
                os.path.basename(self.emulator_start_cmd),
            )

        # 游戏
        object.__setattr__(self, 'app_name', self.game_app.app_name)

        # 设置plan_root
        object.__setattr__(self, 'default_plan_root', os.path.join(DATA_ROOT, 'plans'))
        if self.plan_root is None:
            local_plan_root = os.path.join(os.getcwd(), 'plans')
            if os.path.exists(local_plan_root):
                object.__setattr__(self, 'plan_root', local_plan_root)
                print(f'使用脚本运行目录的plans: {local_plan_root}')
            else:
                object.__setattr__(self, 'plan_root', self.default_plan_root)
                print(f'使用默认plans: {self.default_plan_root}')

        # 加载舰船名文件
        object.__setattr__(self, 'default_ship_name_file', os.path.join(OCR_ROOT, 'ship_name.yaml'))
        if self.ship_name_file is None:
            local_ship_name_file = os.path.join(os.getcwd(), 'ship_name.yaml')
            if os.path.exists(local_ship_name_file):
                object.__setattr__(self, 'ship_name_file', local_ship_name_file)
                print(f'使用本地ship_name: {local_ship_name_file}')
            else:
                object.__setattr__(self, 'ship_name_file', self.default_ship_name_file)
                print(f'使用默认ship_name: {self.default_ship_name_file}')

        # 日志
        object.__setattr__(
            self,
            'log_dir',
            os.path.join(self.log_root, datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')),
        )


@dataclass(frozen=True)
class FightConfig(BaseConfig):
    chapter: int | str = 1
    """章节号"""
    map: int | str = 1
    """地图号"""
    fleet_id: int = 1
    """舰队号"""
    fleet: list[str] | None = None
    """舰队成员 (按名字区分), 为 None 则不变
        fleet: ["", "吹雪", "明斯克", "胡德", "赤诚", "",]
        填 "" 或者不填则代表该位置无舰船 0 号位(第一个)填 "" 占位
        上例的舰队最终为: 吹雪 明斯克 胡德 赤诚 5号位无舰船 6号位无舰船
        现有策略填写时用的作者的船舱名, 记得修改
    """
    repair_mode: RepairMode | list[RepairMode] = RepairMode.severe_damage
    """修理方案。1:中破就修 2:只修大破；也可以用列表指定6个位置不同修理方案"""
    selected_nodes: list[str] = field(default_factory=list)
    """选择要打的节点，白名单模式，一旦到达不要的节点就SL"""
    fight_condition: FightCondition = FightCondition.aim
    """战况选择。1.稳步前进 2.火力万岁 3.小心翼翼 4.瞄准 5.搜索阵型"""

    # 活动专属
    from_alpha: bool = True
    """入口选择。True 为从 alpha 入口进入, False 为从 beta 入口进入"""

    def __post_init__(self) -> None:
        if isinstance(self.repair_mode, list):
            object.__setattr__(self, 'repair_mode', [RepairMode(r) for r in self.repair_mode])
        else:
            object.__setattr__(
                self,
                'repair_mode',
                [RepairMode(self.repair_mode) for _ in range(6)],
            )

        object.__setattr__(self, 'fight_condition', FightCondition(self.fight_condition))


@dataclass(frozen=True)
class BattleConfig(FightConfig):
    repair_mode: RepairMode | list[RepairMode] = RepairMode.moderate_damage
    """修理方案。1:中破就修 2:只修大破；也可以用列表指定6个位置不同修理方案"""


@dataclass(frozen=True)
class ExerciseConfig(FightConfig):
    selected_nodes: list[str] = field(default_factory=lambda: ['player', 'robot'])
    """仅使用默认值"""
    discard: bool = False
    """TODO: 检查逻辑"""
    exercise_times: int = 4
    """最大玩家演习次数"""
    robot: bool = True
    """是否打机器人"""
    max_refresh_times: int = 2
    """最大刷新次数"""


# @dataclass(frozen=True)
# class EnemyRule(BaseConfig):
#     condition: str
#     """触发条件"""
#     action: SearchEnemyAction | Formation
#     """进行动作，或者选择阵型"""

#     def __post_init__(self) -> None:
#         if isinstance(self.action, int):
#             object.__setattr__(self, 'action', Formation(self.action))
#         else:
#             object.__setattr__(self, 'action', SearchEnemyAction(self.action))


@dataclass(frozen=True)
class NodeConfig(BaseConfig):
    # 索敌阶段
    long_missile_support: bool = False
    """（当存在时）是否开启导巡的远程导弹支援"""
    detour: bool = False
    """（当存在时）是否进行迂回"""
    enemy_rules: list[str] = field(default_factory=list)
    """每条按照 ["(类型 符号 数量) and/or ()", "操作"] 的字符串形式给出，从上到下执行第一条符合的命令。
    类型：舰船类型；符号：[>=、<=、>、<、==、!=]；数量：数字；
    操作：[retreat、detour、阵型数字(1-5)]
    如果在这里做了阵型选择决定，则优先级将高过下面的阵型选择。例子：["(SS >= 2)", 5]
    and 的优先级高于 or"""

    enemy_formation_rules: list[str] = field(default_factory=list)
    """同enemy_rules. overrides enemy_rules. e.g. [单横, retreat]"""

    # 阵型选择阶段
    SL_when_spot_enemy_fails: bool = False
    """索敌失败时是否SL"""
    SL_when_detour_fails: bool = True
    """迂回失败是否退出"""
    SL_when_enter_fight: bool = False
    """进入战斗是否退出"""
    formation: Formation = Formation.double_column
    """阵型选择"""
    formation_when_spot_enemy_fails: Formation | None = None
    """索敌失败时阵型选择"""

    # 夜战, 前进阶段
    night: bool = False
    """是否夜战"""
    proceed: bool = True
    """是否前进"""
    proceed_stop: RepairMode | list[RepairMode] = RepairMode.severe_damage
    """达到指定破损状态时结束。1:中破 2:大破；也可以用列表指定6个位置不同"""

    def __post_init__(self) -> None:
        if self.enemy_rules is None:
            object.__setattr__(self, 'enemy_rules', [])
        object.__setattr__(self, 'formation', Formation(self.formation))
        if self.formation_when_spot_enemy_fails is not None:
            object.__setattr__(
                self,
                'formation_when_spot_enemy_fails',
                Formation(self.formation_when_spot_enemy_fails),
            )
        if isinstance(self.proceed_stop, list):
            object.__setattr__(self, 'proceed_stop', [RepairMode(r) for r in self.proceed_stop])
        else:
            object.__setattr__(
                self,
                'proceed_stop',
                [RepairMode(self.proceed_stop) for _ in range(6)],
            )


ATTRIBUTE_RECURSIVE = {
    'daily_automation': DailyAutomationConfig,
    'decisive_battle': DecisiveBattleConfig,
    # {'enemy_rules': EnemyRule},
}
ATTRIBUTE_IGNORE = {
    'node_defaults',
    'node_args',
}
